// File: ClassifierTest.h
// Purpose: Definition of the ClassifierTest class and its factory class
// Remarks: This class is an abstract base class.
// Author: Yihua Chen (yhchen@u.washington.edu)
// Last modified: 11/15/2007 3:04:00 PM

#ifndef IDL_CLASSIFIERTEST_H
#define IDL_CLASSIFIERTEST_H

#include <map>
#include <string>
#include <fstream>
#include "ConfigParser.h"
#include "RealMatrix.h"
#include "SimData.h"

namespace idl
{
	/** Abstract base class for testing a specific classifier on similarity data. */
	class ClassifierTest
	{
	public:
		ClassifierTest() : testCount(0) {}	/**< Constructor. */
		virtual ~ClassifierTest() {}		/**< Destructor. */

		/** Clear the records of the tests on all available parameters. */
		virtual void reset();

		/**
		 * Clear the records of the tests on all available parameters.
		 * @param[in]	numParameterSets	number of the parameter sets to test on
		 */
		virtual void reset(unsigned int numParameterSets);

		/**
		 * Clear the results with the best parameters.
		 * @attention	This is a pure virtual function.
		 */
		virtual void resetBestParaResults() = 0;

		/**
		 * Select the best parameters based on the previous tests.
		 * @attention	This is a pure virtual function.
		 */
		virtual void selectBestPara() = 0;

		/**
		 * Configure the test for a specific classifier.
		 * @param[in]	config	ConfigParser object containing user configurations
		 * @param[in]	secIdx	section index
		 * @return		true if success; false if failure
		 * @attention	This is a pure virtual function.
		 */
		virtual bool configure(ConfigParser &config, int secIdx) = 0;

		/**
		 * Test the classifier on all available parameters.
		 * @param[in]	data	similarity data to test on
		 * @param[in]	logFile	log file stream
		 * @return		true if success; false if failure
		 * @attention	This is a pure virtual function.
		 */
		virtual bool test(const SimData &data, std::ofstream &logFile) = 0;

		/**
		 * Test the classifier with the best parameters.
		 * @param[in]	data	similarity data to test on
		 * @param[in]	logFile	log file stream
		 * @return		true if success; false if failure
		 * @attention	This is a pure virtual function.
		 */
		virtual bool testBestPara(const SimData &data, std::ofstream &logFile) = 0;

		/**
		 * Report the test results of all available parameters.
		 * @param[in]	file	output file stream
		 * @param[in]	verbose	flag for verbose/concise output: true for verbose;
		 *						false for concise (false by default)
		 * @remarks		For concise output, only the best result and its corresponding
		 *				parameters are reported. For verbose output, the results of all
		 *				available parameter sets are reported.
		 * @attention	This is a pure virtual function.
		 */
		virtual void reportResults(std::ofstream &file, bool verbose = false) = 0;

		/**
		 * Report the test results with the best parameters.
		 * @param[in]	file	output file stream
		 * @param[in]	verbose	flag for verbose/concise output: true for verbose;
		 *						false for concise (false by default)
		 * @attention	This is a pure virtual function.
		 */
		virtual void reportBestParaResults(std::ofstream &file, bool verbose = false) = 0;

		/**
		 * Summarize the results of all the tests with the best parameters.
		 * @param[in]	file	output file stream
		 * @param[in]	verbose	flag for verbose/concise output: true for verbose;
		 *						false for concise (false by default)
		 * @attention	This is a pure virtual function.
		 */
		virtual void reportBestParaSummary(std::ofstream &file, bool verbose = false) = 0;

	protected:
		/**
		 * Get the cross-validation error of a parameter set.
		 * @param[in]	idx	index of the parameter set
		 * @return		cross-validation error
		 */
		double getError(unsigned int idx) const;

		/**
		 * Get the standard deviation of the cross-validation error.
		 * @param[in]	idx	index of the parameter set
		 * @return		standard deviation of the cross-validation error
		 */
		double getErrorStdDev(unsigned int idx) const;

		/**
		 * Get the information about the minimum cross-validation error.
		 * @param[out]	idx		index of the parameter set achieving minimum error
		 * @param[out]	error	value of the minimum error
		 * @param[out]	stddev	standard deviation of the error
		 */
		void getMinErrorInfo(unsigned int &idx, double &error, double &stddev) const;

	protected:
		unsigned int testCount; /**< Number of the tests performed. */

		/**
		 * Vector to store the sum of errors.
		 * Vector to store the sum of classification errors for each parameter set.
		 * @remarks	This is used for calculating the mean of the classification errors
		 *			for each parameter set.
		 */
		RealMatrix errorSum;

		/**
		 * Vector to store the sum of squared errors.
		 * Vector to store the sum of squared classification erros for each parameter set.
		 * @remarks	This is used for calculating the variance of the classification errors
		 *			for each parameter set.
		 */
		RealMatrix errorSumOfSquares;
	};

	/**
	 * Factory class for ClassifierTest.
	 * Class for instantiating classes derived from the ClassifierTest class.
	 * @remarks	The factory method pattern is applied here.
	 * @see	ClassifierTest
	 */
	class ClassifierTestFactory
	{
	private:
		/**
		 * Type definition for the ClassifierTest creators.
		 * Type definition for pointers to the functions that create objects
		 * of the classes derived from the ClassifierTest class.
		 */
		typedef ClassifierTest *(*CreatorType)();

		/**
		 * Type definition for the ClassifierTest creator map.
		 * Type definition for the map that associates the ID of a class
		 * derived from the ClassifierTest class and its creator.
		 */
		typedef std::map<std::string, CreatorType> MapType;

		/**
		 * Get the map for the ClassifierTest creators.
		 * @return	Reference to the ClassifierTest creator map
		 */
		static MapType & getCreatorMap();

	public:
		/**
		 * Register a ClassifierTest creator.
		 * @param[in]	id		class ID
		 * @param[in]	creator	class instantiator
		 * @return		true if success; false if the same class ID already exists
		 */
		static bool registerCreator(const std::string &id, CreatorType creator);

		/**
		 * Create an instance of the specified ClassifierTest subclass.
		 * @param[in]	id	class ID
		 * @return		pointer to the instance (NULL if failure)
		 */
		static ClassifierTest *create(const std::string &id);
	};
}

#endif // IDL_CLASSIFIERTEST_H
