// File: LimeSolver.cpp
// Purpose: Implementation of the LimeSolver class
// Author: Yihua Chen (yhchen@u.washington.edu)
// Last modified: 7/25/2007 10:33AM

#include <cmath>
#include <cstdlib>
#include <cstdio>
#include <cstring>
#include <iostream>
#include "mosek.h" // MOSEK header file
#include "LimeSolver.h"

using namespace std;

namespace idl
{
	// Initialize the number of equality constraints
	const int LimeSolver::NUM_EQ_CON = 1;

	MSKboundkeye LimeSolver::bkc[1] = {MSK_BK_FX}; // Type of constraint: fixed (equality)

	double LimeSolver::blc[1] = {1.0}; // Lower bound for the equality constraint
	double LimeSolver::buc[1] = {1.0}; // Upper bound for the equality constraint

	int *LimeSolver::ptrb = NULL;
	int *LimeSolver::ptre = NULL;
	int *LimeSolver::sub  = NULL;

	MSKboundkeye *LimeSolver::bkx  = NULL;

	double *LimeSolver::blx = NULL;
	double *LimeSolver::bux = NULL;
	double *LimeSolver::val = NULL;
	double *LimeSolver::sol = NULL;

	int LimeSolver::numVariables = 0;
	int LimeSolver::numHesNz = 0;

	double *LimeSolver::Q = NULL;
	double *LimeSolver::q = NULL;
	double LimeSolver::lambda = 0.0;

	// Compute f(x), the nonlinear part of the objective function
	// f(x) = 1/2 * x' * Q * x - lambda * H(x)
	double LimeSolver::f(double *x)
	{
		double r1 = 0.0;

		for (int j = 0; j < numVariables; ++j)
		{
			double y = 0.0;

			for (int i = 0; i < numVariables; ++i)
			{
				y += x[i] * Q[i * numVariables + j];
			}

			r1 += y * x[j];
		}

		r1 /= 2; // r1 = 1/2 * x' * Q * x

		double r2 = 0.0;

		for (int i = 0; i < numVariables; ++i)
		{
			if (x[i] > 0.0)
			{
				r2 += x[i] * log(x[i]);
			}
		}

		r2 *= lambda; // r2 = -lambda * H(x)

		return (r1 + r2);
	}

	// Compute the gradient of f(x)
	// grad f(x) = Q * x + lambda * [log(x_1) + 1, ..., log(x_n) + 1]'
	double LimeSolver::f_grad(double *x, int i)
	{
		double r = 0.0;

		for (int j = 0; j < numVariables; ++j)
		{
			r += Q[i * numVariables + j] * x[j];
		}

		r += lambda * (log(x[i]) + 1);

		return r;
	}

	// Compute the Hessian of f(x)
	// H f(x) = Q_ii + lambda / x_i, i = j; Q_ij, i != j
	double LimeSolver::f_hessian(double *x, int i, int j)
	{
		double r = Q[i * numVariables + j];

		return (i != j) ? r : (r + lambda / x[i]);
	}

	// Print the messages from MOSEK on the screen
	// Remarks: This is a callback function for MOSEK.
	void MSKAPI LimeSolver::mosekPrintLog(void *handle, char str[])
	{
#ifdef _DEBUG
		printf("%s", str);
#endif // _DEBUG
	}

	// Inform MOSEK of the problem structure and sparsity
	// Remarks: This is a callback function for MOSEK.
	int MSKAPI LimeSolver::mosekNlSpar(void *nlhandle,
									   int  *numgrdobjnz,
									   int  *grdobjsub,
									   int  i,
									   int  *convali,
									   int  *grdconinz,
									   int  *grdconisub,
									   int  yo,
									   int  numycnz,
									   int  ycsub[],
									   int  maxnumhesnz,
									   int  *numhesnz,
									   int  *hessubi,
									   int  *hessubj)
	{
		if (numgrdobjnz != NULL)
		{
			// Number of non-zero elements in the gradient of f(x)
			*numgrdobjnz = numVariables;

			if (grdobjsub != NULL)
			{
				// Positions of non-zero elements in the gradient of f(x)
				for (int i = 0; i < numVariables; ++i)
				{
					grdobjsub[i] = i; 
				}
			}
		}

		if (convali != NULL)
		{
			*convali = 0; // Zero because no nonlinear constraint
		}

		if (grdconinz != NULL)
		{
			*grdconinz = 0; // Zero because no nonlinear constraint
		}

		if (numhesnz != NULL)
		{
			if (yo != 0)
			{
				// Set the number of non-zero elements in the lower
				// triangle of the Hessian matrix of f(x)
				*numhesnz = numHesNz;
			}
			else
			{
				*numhesnz = 0;
			}
		}

		if (maxnumhesnz != 0)
		{
			if (maxnumhesnz < numHesNz)
			{
				// No enough space to store the Hessian of f(x)
				return MSK_RES_ERR_INTERNAL;
			}
			else
			{
				if (yo != 0 && hessubi != NULL && hessubj != NULL)
				{
					int k = 0;

					// Positions of non-zero elements in the lower
					// triangle of the Hessian of f(x)
					for (int i = 0; i < numVariables; ++i)
					{
						for (int j = 0; j <= i; ++j)
						{
							if (j == i || Q[i * numVariables + j] != 0.0)
							{
								hessubi[k] = i;
								hessubj[k++] = j;
							}
						}
					}
				}
			}
		}

		return MSK_RES_OK;
	}

	// Evaluate the nonlinear function (including its gradient and Hessian)
	// and return the requested information to MOSEK
	// Remarks: This is a callback function for MOSEK.
	int MSKAPI LimeSolver::mosekNlEval(void		*nlhandle,
									   double	*xx,
									   double	yo,
									   double	*yc,
									   double	*objval,
									   int		*numgrdobjnz,
									   int		*grdobjsub,
									   double	*grdobjval,
									   int		numi,
									   int		*subi,
									   double	*conval,
									   int		*grdconptrb,
									   int		*grdconptre,
									   int		*grdconsub,
									   double	*grdconval,
									   double	*grdlag,
									   int		maxnumhesnz,
									   int		*numhesnz,
									   int		*hessubi,
									   int		*hessubj,
									   double	*hesval)
	{
		// Evaluate the nonlinear function f(x)
		if (objval != NULL)
		{
			*objval = f(xx);
		}

		// Evaluate the gradient of f(x)
		if (numgrdobjnz != NULL)
		{
			// Number of non-zero elements in the gradient of f(x)
			*numgrdobjnz = numVariables;

			for (int i = 0; i < numVariables; ++i)
			{
				// Positions of non-zero elements in the gradient of f(x)
				grdobjsub[i] = i;

				// Values of non-zero elements in the gradient of f(x)
				grdobjval[i] = f_grad(xx, i);
			}
		}

		if (conval != NULL)
		{
			// Zero because no nonlinear constraint
			for (int k = 0; k < numi; ++k)
			{
				conval[k] = 0.0;
			}
		}

		if (grdlag != NULL)
		{
			// Gradient of the Lagrangian function yo * f(x)
			for (int i = 0; i < numVariables; ++i)
			{
				grdlag[i] = yo * f_grad(xx, i);
			}
		}

		// Evaluate the Hessian of f(x)
		if (maxnumhesnz != 0)
		{
			if (yo == 0.0)
			{
				*numhesnz = 0;
			}
			else
			{
				if (maxnumhesnz < numHesNz)
				{
					// No enough space to store the Hessian of f(x)
					return MSK_RES_ERR_INTERNAL;
				}
				else
				{
					// Number of non-zero elements in the lower
					// triangle of the Hessian of f(x)
					*numhesnz = numHesNz;

					if (hessubi != NULL && hessubj != NULL && hesval != NULL)
					{
						int k = 0;

						for (int i = 0; i < numVariables; ++i)
						{
							for (int j = 0; j <= i; ++j)
							{
								if (j == i || Q[i * numVariables + j] != 0.0)
								{
									// Positions of non-zero elements in the lower
									// triangle of the Hessian of f(x)
									hessubi[k] = i;
									hessubj[k] = j;

									// Values of non-zero elements in the lower
									// triangle of the Hessian of f(x)
									hesval[k++] = f_hessian(xx, i, j);
								}
							}
						}
					}
				}
			}
		}

		return MSK_RES_OK;
	}

	// Clear the information about the optimization problem
	void LimeSolver::clearProbInfo()
	{
		numVariables = 0;

		delete [] ptrb;
		delete [] ptre;
		delete [] sub;
		delete [] bkx;
		delete [] blx;
		delete [] bux;
		delete [] val;
		delete [] sol;
	}

	// Set the information about the optimization problem
	bool LimeSolver::setProbInfo(int numNeighbors)
	{
		if (numNeighbors < 0)
		{
			cerr << "Error: Number of neighbors must be positive" << endl;
			return false;
		}

		if (numVariables != numNeighbors)
		{
			clearProbInfo();

			numVariables = numNeighbors;

			ptrb = new int[numVariables];
			ptre = new int[numVariables];
			sub  = new int[numVariables];

			bkx  = new MSKboundkeye[numVariables];

			blx = new double[numVariables];
			bux = new double[numVariables];
			val = new double[numVariables];
			sol = new double[numVariables];

			for (int i = 0; i < numVariables; ++i)
			{
				ptrb[i] = i;			// Pointer to the first element in columns of A
				ptre[i] = i+1;			// Pointer to the last element + 1 in columns of A
				sub[i]  = 0;			// Constraint subscripts
				val[i]  = 1.0;			// Constraint values (elements of A)
				bkx[i]  = MSK_BK_LO;	// Type of constraint: lower bound
				blx[i]  = 0.0;			// Lower bound of the inequality constraint
				bux[i]  = MSK_INFINITY;	// Upper bound of the inequality constraint
			}
		}

		return true;
	}

	// Clear the parameters of the optimization problem
	void LimeSolver::clearProbPara()
	{
		Q = NULL;
		q = NULL;
	}

	// Set the parameters of the optimization problem
	bool LimeSolver::setProbPara(double *QMtx, double *qVec, double lambdaVal)
	{
		if (NULL == QMtx)
		{
			cerr << "Error: Invalid pointer to matrix Q" << endl;
			return false;
		}

		if (NULL == qVec)
		{
			cerr << "Error: Invalid pointer to vector q" << endl;
			return false;
		}

		if (lambdaVal <= 0.0)
		{
			cerr << "Error: Value of lambda must be positive" << endl;
			return false;
		}

		Q = QMtx;
		q = qVec;
		lambda = lambdaVal;

		// Count the number of non-zero elements in the lower triangle
		// of the Hessian matrix of f(x), which is equal to the number
		// of non-zero elements in the lower triangle of Q
		numHesNz = numVariables;

		for (int i = 0; i < numVariables; ++i)
		{
			for (int j = 0; j < i; ++j)
			{
				if (0.0 != Q[i * numVariables + j])
				{
					++numHesNz;
				}
			}
		}

		return true;
	}

#ifdef _DEBUG
	// Print the parameters of the optimization problem on the screen
	void LimeSolver::printProbPara()
	{
		cout << "numVariables = " << numVariables << endl;
		
		cout << "Q =" << endl;
		for (int i = 0; i < numVariables; ++i)
		{
			for (int j = 0; j < numVariables; ++j)
			{
				cout << Q[i * numVariables + j] << '\t';
			}
			cout << endl;
		}

		cout << "numHesNz = " << numHesNz << endl;

		cout << "q =" << endl;
		for (int i = 0; i < numVariables; ++i)
		{
			cout << q[i] << endl;
		}

		cout << "lambda = " << lambda << endl;
	}
#endif // _DEBUG

	// Solve the convex optimization problem for LIME
	bool LimeSolver::solve(int		numNeighbors,
						   double	*QMtx,
						   double	*qVec,
						   double	lambdaVal,
						   double	*weights)
	{
		if (NULL == weights)
		{
			cerr << "Error: Invalid pointer to the output buffer" << endl;
			return false;
		}

		if (!setProbInfo(numNeighbors)) return false;
		if (!setProbPara(QMtx, qVec, lambdaVal)) return false;

#ifdef _DEBUG
		printProbPara(); // Print the parameters of the problem
#endif // _DEBUG

		///////////////////////////////////////////////////////////////////////
		// Solve the optimization problem by MOSEK
		///////////////////////////////////////////////////////////////////////

		MSKenv_t env; // MOSEK environment

		// Create a new MOSEK environment
		if (MSK_RES_OK != MSK_makeenv(&env, NULL, NULL, NULL, NULL))
		{
			cerr << "Error: Cannot create a new MOSEK environment" << endl;
			return false;
		}
		else
		{
			// Link MOSEK environment log stream to the standard output
			MSK_linkfunctoenvstream(env, MSK_STREAM_LOG, NULL, mosekPrintLog);
		}

		// Initialize the MOSEK environment
		if (MSK_RES_OK != MSK_initenv(env))
		{
			cerr << "Error: Cannot initialize the MOSEK environment" << endl;
			return false;
		}

		MSKtask_t task; // MOSEK task

		// Create an optimization task
		if (MSK_RES_OK != MSK_maketask(env, NUM_EQ_CON, numVariables, &task))
		{
			cerr << "Error: Cannot create an optimization task" << endl;
			return false;
		}
		else
		{
			// Link MOSEK task log stream to the standard output
			MSK_linkfunctotaskstream(task, MSK_STREAM_LOG, NULL, mosekPrintLog);
		}

		// Set the linear part of the optimization problem
		if (MSK_RES_OK != MSK_inputdata(
			task,			// A MOSEK optimization task
			NUM_EQ_CON,		// Maximum number of constraints in the optimization task
			numVariables,	// Maximum number of variables in the optimization task
			NUM_EQ_CON,		// Number of constraints
			numVariables,	// Number of variables
			q,				// Linear term in the objective function
			0.0,			// Fixed term in the objective function
			ptrb,			// Pointer to the first element in columns of A
			ptre,			// Pointer to the last element + 1 in columns of A
			sub,			// Constraint subscripts
			val,			// Constraint values (elements of A)
			bkc,			// Bound keys for the constraints
			blc,			// Lower bounds for the constraints
			buc,			// Upper bounds for the constraints
			bkx,			// Bound keys for the variables
			blx,			// Lower bounds for the variables
			bux				// Upper bounds for the variables
			))
		{
			cerr << "Error: Cannot set the linear part of the optimization problem" << endl;
			return false;
		}

		// Set the nonlinear part of the optimization problem
		if (MSK_RES_OK != MSK_putnlfunc(task, NULL, mosekNlSpar, mosekNlEval))
		{
			cerr << "Error: Cannot set the nonlinear part of the optimization problem" << endl;
			return false;
		}

		// Solve the optimization problem
		if (MSK_RES_OK != MSK_optimize(task))
		{
			cerr << "Error: Cannot solve the optimization problem" << endl;
			return false;
		}

		// Get the solution
		MSK_getsolutionslice(task, MSK_SOL_ITR, MSK_SOL_ITEM_XX, 0, numVariables, sol);

#ifdef _DEBUG
		// Print a short summary of the solution
		MSK_solutionsummary(task, MSK_STREAM_LOG);

		// Print the solution
		cout << "sol =" << endl;
		for (int i = 0; i < numVariables; ++i)
		{
			cout << sol[i] << endl;
		}
#endif // _DEBUG

		memcpy(weights, sol, numVariables * sizeof(double)); // Copy the solution

		///////////////////////////////////////////////////////////////////////////
		// Clean up the environment
		///////////////////////////////////////////////////////////////////////////

		MSK_deletetask(&task);
		MSK_deleteenv(&env);

		clearProbPara();

		return true;
	}
}
