// File: LocalSDA.h
// Purpose: Definition of the LocalSDA class
// Remarks: This class is designed for Local SDA classification.
// Author: Ingrid(ingrid@hkn.eecs.berkeley.edu)
// Modified by Peter Sadowski peterjsadowski@gmail.com

#ifndef IDL_LocalSDA_H
#define IDL_LocalSDA_H

#include <vector>
#include <fstream>
#include "cxcore.h"	// OpenCV CXCORE header file
#include "ml.h"		// OpenCV MLL header file
#include "RealMatrix.h"
#include "Classifier.h"
#include "SimData.h"

namespace idl
{
	/**
	 * Class for the Local SDA classifier.
	 * @remarks	This class is derived from the Classifier class.
	 * @see		Classifier
	 */
	class LocalSDA : public Classifier
	{
	public:
		/** Enumeration type for embedding method. */

		/** Structure type for the Local SDA parameters. */
		struct ParaType
		{
			unsigned int neighborhoodSize;	/**< Neighborhood size. */
		};

		/** Default constructor. */
		LocalSDA()
		{
			parameters.neighborhoodSize = 1;
		}

		/**
		 * Constructor.
		 * @param	para	parameters for the Local SDA classifier
		 */
		LocalSDA(const ParaType &para)
		{
			parameters = para;
		}

		~LocalSDA() {}	/**< Destructor. */

		/**
		 * Set the neighborhood size.
		 * @param[in]	k	neighborhood size
		 */
		void setNeighborhoodSize(unsigned int k)
		{
			if (k > 0)
			{
				parameters.neighborhoodSize = k;
			}
		}

		/**
		 * Set the parameters for the Local SDA classifier.
		 * @param[in]	para	parameters for the Local SDA classifier
		 */
		void setPara(const ParaType &para)
		{
			parameters = para;
		}

		/**
		 * Classify the test samples of the similarity data set.
		 * @param[in]	data		similarity data set
		 * @param[out]	predictedClassLabels		vector to store the estimated class labels
		 * @return		true if success; false if failure
		 */
		bool classify(const SimData &data, std::vector<int> &predictedClassLabels);
		bool classify(const SimData &data, std::vector<int> &predictedClassLabels, RealMatrix &classProbMatrix);

	private:
		ParaType parameters;	/**< Parameters for the Local SDA classifier. */
	};
}

#endif // IDL_LocalSDA_H
