// File: LocalSDABN.h
// Purpose: Definition of the LocalSDABN class
// Remarks: This class is designed for Local SDA classification with Bayesian Neighborhoods
// Author: Ingrid(ingrid@hkn.eecs.berkeley.edu)
// Last modified: 7/11/2009 11:45:18 AM

#ifndef IDL_LocalSDABN_H
#define IDL_LocalSDABN_H

#include <vector>
#include <fstream>
#include "cxcore.h"	// OpenCV CXCORE header file
#include "ml.h"		// OpenCV MLL header file
#include "RealMatrix.h"
#include "Classifier.h"
#include "SimData.h"
#include "LocalSDA.h"


namespace idl
{
	/**
	 * Class for the Local SDA classifier.
	 * @remarks	This class is derived from the Classifier class.
	 * @see		Classifier
	 */
	class LocalSDABN : public LocalSDA
	{
	public:

		/** Structure type for the Local SDA parameters. */
		struct ParaType
		{
			std::vector<unsigned int> neighborhoodSizes;	/**< Neighborhood size. */
			double regParam;				/**< Regularization parameter  */
		};

		/** Default constructor. */
		LocalSDABN()
		{
			//parameters.neighborhoodSize = 1;
			//parameters.regParam = 1.0;
		}

		/**
		 * Constructor.
		 * @param	para	parameters for the Local SDA classifier
		 */
		LocalSDABN(const ParaType &para)
		{
			parameters = para;
		}

		~LocalSDABN() {}	/**< Destructor. */

		/**
		 * Set the neighborhood size.
		 * @param[in]	k	neighborhood size
		 */
		void setNeighborhoodSizes(const std::vector<unsigned int> nbSizes)
		{
			parameters.neighborhoodSizes = nbSizes;
		}

		/**
		 * Set the parameters for the Local SDA classifier.
		 * @param[in]	para	parameters for the Local SDA classifier
		 */
		void setPara(const ParaType &para)
		{
			parameters = para;
		}

		/**
		 * Classify the test samples of the similarity data set.
		 * @param[in]	data		similarity data set
		 * @param[out]	results		vector to store the estimated class labels
		 * @return		true if success; false if failure
		 */
		bool classify(const SimData &data, std::vector<int> &results);
		bool classify(const SimData &data, std::vector<int> &results, RealMatrix &probMatrix);

		//bool getProbabilities(SimData &data, unsigned int testIdx, RealMatrix &probabilities);

	private: 
		int maxLabel(const std::vector<int> &labels, const RealMatrix &probabilities);

	private:

		ParaType parameters;	/**< Parameters for the Local SDA classifier. */
		//LocalSDABNModel lsdaModel;	/**< The model for this test sample. */
	};
}

#endif // IDL_LocalSDABN_H
