// File: LocalSDABNTest.h
// Purpose: Definition of the LocalSDABNTest class
// Author: Ingrid(ingrid@hkn.eecs.berkeley.edu)
// Last modified: 7/11/2009 11:45:18 AM

#ifndef IDL_LOCALSDABNTEST_H
#define IDL_LOCALSDABNTEST_H

#include <vector>
#include <string>
#include <fstream>
#include "ClassifierTest.h"
#include "LocalSDABN.h"

namespace idl
{
	/**
	 * Class for testing a weighted k-NN classifier on similarity data.
	 * An adapter class for the LocalSDA class to enable testing on multiple
	 * sets of parameters.
	 * @remarks	This class is derived from the ClassifierTest class.
	 * @see		ClassifierTest
	 * @see		ClassifierTestFactory
	 * @see		LocalSDA
	 */
	class LocalSDABNTest : public ClassifierTest
	{
	public:
		LocalSDABNTest() : ClassifierTest() {}	/**< Constructor. */

		LocalSDABNTest(std::vector<unsigned int> nbrSizes) : ClassifierTest() 
		{
			neighborhoodSizes = nbrSizes;
		}
		~LocalSDABNTest();						/**< Destructor. */

		/** Clear the results with the best parameters. */
		void resetBestParaResults();

		/** Select the best parameters based on the previous tests. */
		void selectBestPara();

		/**
		 * Configure the test for a weighted k-NN classifier.
		 * @param[in]	config		ConfigParser object containing user configurations
		 * @param[in]	secIdx		section index
		 * @return		true if success; false if failure
		 */
		bool configure(ConfigParser &config, int secIdx);

		/**
		 * Test the weighted k-NN classifier on all available parameters.
		 * @param[in]	data		similarity data to test on
		 * @param[in]	logFile		log file stream
		 * @return		true if success; false if failure
		 */
		bool test(const SimData &data, std::ofstream &logFile);

		/**
		 * Test the weighted k-NN classifier with the best parameters.
		 * @param[in]	data		similarity data to test on
		 * @param[in]	logFile		log file stream
		 * @return		true if success; false if failure
		 */
		bool testBestPara(const SimData &data, std::ofstream &logFile);

		/**
		 * Report the test results of all available parameters.
		 * @param[in]	file		output file stream
		 * @param[in]	verbose		flag for verbose/concise output: true for verbose;
		 *							false for concise (false by default)
		 * @remarks		For concise output, only the best result and its corresponding
		 *				parameters are reported. For verbose output, the results of all
		 *				available parameter sets are reported.
		 */
		void reportResults(std::ofstream &file, bool verbose = false);

		/**
		 * Report the test results with the best parameters.
		 * @param[in]	file		output file stream
		 * @param[in]	verbose		flag for verbose/concise output: true for verbose;
		 *							false for concise (false by default)
		 */
		void reportBestParaResults(std::ofstream &file, bool verbose = false);

		/**
		 * Summarize the results of all the tests with the best parameters.
		 * @param[in]	file		output file stream
		 * @param[in]	verbose		flag for verbose/concise output: true for verbose;
		 *							false for concise (false by default)
		 */
		void reportBestParaSummary(std::ofstream &file, bool verbose = false);

	private:
		/**
		 * Check the legitimacy of the regularization parameters.
		 * @return	true if all the regularization parameters are within the legal range;
		 *			false if at least one regularization parameter is out of the legal range
		 */
		//bool checkLambdas();

		/**
		 * Check the legitimacy of the neighborhood sizes.
		 * @return	true if all the neighborhood sizes are within the legal range;
		 *			false if at least one neighborhood size is out of the legal range
		 */
		bool checkNeighborhoodSizes();

	private:
		LocalSDABN lsda; /**< Weighted k-NN classifier. */
		//std::vector<double> regParams; /**< Vector of regularization parameters. */
		std::vector<unsigned int> neighborhoodSizes; /**< Vector of neighborhood sizes. */

		/** Structure type for test results with the best parameters. */
		struct ResultType
		{
			double cvError;					/**< Cross-validation error. */
			double cvErrorStdDev;			/**< Standard deviation of the cross-validation error. */
			double testError;				/**< Test error. */
		};

		std::vector<ResultType> resultsBestPara; /**< Test results with the best parameters. */

	private:
		/**
		 * Create an object of the LocalSDABNTest class.
		 * @return	pointer to a new object of the LocalSDABNTest class
		 * @remarks	This is the creator function to be registered with the factory.
		 */
		static ClassifierTest *create()
		{
			return new LocalSDABNTest;
		}
		static const std::string _ID;	/**< Class ID ("kNN") */
		static const bool _REGISTERED;	/**< Flag for creator registration */
	};
}

#endif // IDL_LOCALSDATEST_H
