// File: LocalSDAModel.cpp
// Purpose: Implementation of the LocalSDAModel class
// Author: Ingrid Liu (ingrid@hkn.eecs.berkeley.edu), Peter Sadowski (peterjsadowski@gmail.com)
// Last modified: 

#include <cmath>
#include <vector>
#include <map>
#include <algorithm>
#include "LocalSDAModel.h"
#include "RealMatrix.h"
#include "SimData.h"
#include "brent.h"
#define INF 1.7E308;


using namespace std;

namespace idl 
{
	LocalSDAModel::LocalSDAModel() 
	{ }

	// Destructor
	LocalSDAModel::~LocalSDAModel()
	{ }
	
	bool LocalSDAModel::reset() 
	{
		samplesByLabel.clear();
		uniqueClassLabelsLocal.clear();
		actualToIndexedLabels.clear();
		centroidIdx.clear();
		centroidsToTestSim.clear();
		lambdas.clear();
		gammas.clear();
		classPriors.clear();
		anyOneClassFlag = false;
		onlyOneClassFlag = false;
		oneNeighbor = false;
		allSimsZeroFlag = false;
		return true;
	}

	// Setup the local model.
	void LocalSDAModel::initialize(const RealMatrix &simMat, const vector<int> &orig_labels, const RealMatrix &simTest, const RealMatrix &domainVector)									
	{
		// Configure Variables:
		int nSamples = simMat.size()[0];
		if( ! (reset()
			&& setUniqueLabels(orig_labels,uniqueClassLabelsLocal,actualToIndexedLabels)
			&& setSamplesByLabel(samplesByLabel, orig_labels)	// Structure to get all train indices from a given class.
			&& setClassCentroids(simMat, centroidIdx)
		    && setCentroidsToTestSim(centroidsToTestSim,simTest)
		    && setPriors(classPriors,nSamples)
		    && setDegenerateFlags(simTest, anyOneClassFlag,onlyOneClassFlag, oneNeighbor, allSimsZeroFlag) ) )
		{  // For code robustness: allow possibility of initialization failure.
			cerr << "Initialization of LocalSDAModel failed" << endl; 
		}
	}

	/* Compute the model for this neighborhood.
	 * Saves lambda and gamma matrices (nClasses*nClasses matrices).  
	 */
	void LocalSDAModel::computeModelParams(const RealMatrix &simMat, const vector<int> &orig_labels, const RealMatrix &simTest, const RealMatrix &domainVector)
	{
		int nClasses = (int)uniqueClassLabelsLocal.size();
		lambdas = RealMatrix(nClasses,nClasses,-10e20);
		gammas = RealMatrix(nClasses,nClasses,1);

		for (vector<int>::const_iterator itrG = uniqueClassLabelsLocal.begin(); 
			itrG != uniqueClassLabelsLocal.end(); ++itrG)
		{
			for (vector<int>::const_iterator itrH = uniqueClassLabelsLocal.begin(); 
			itrH != uniqueClassLabelsLocal.end(); ++itrH)
			{
				
				// simUhG = similarities between centroid H and samples of label G
				RealMatrix simUhG = simMat(centroidIdx[*itrH],samplesByLabel[*itrG]);
				// classMean = E[s(v,u_h) | Y=g]; Y = label(v)
				double classMean = simUhG.sum()/simUhG.size()[1];
				double tmpLambda;
				double tmpGamma;
				// Calculate lambda, gamma.
				if (!getLambdaGamma(domainVector, classMean, tmpLambda, tmpGamma)) {
					cerr << "ERROR: Calculation of lambda, gamma failed." << endl; 
				}
				unsigned int indexG = actualToIndexedLabels[*itrG];
				unsigned int indexH = actualToIndexedLabels[*itrH];
				lambdas(indexG,indexH) = tmpLambda;
				gammas(indexG,indexH) = tmpGamma;
			}
		}
	}

	//////////////////////////////
	// INITIALIZATION SUBFUNCTIONS //
	//////////////////////////////

	/* Set uniqueClassLabelsLocal 
	 * Finds all unique labels and saves labels to uniqueClassLables 
	 */
	bool LocalSDAModel::setUniqueLabels(const vector<int> &labels,
										vector<int> &uniqueClassLabelsLocal,
										map<int,unsigned int> &actualToIndexedLabels)
	{
		map<int,int> seen;
		// Find all unique labels and save in the temporary map "seen".
		for(vector<int>::const_iterator itr = labels.begin();
			itr != labels.end(); ++itr)
		{
			seen[*itr] = 1;
		}
		//Save unique class labels and index them (in sorted order) starting at 0 into uniqueClassLabelsLocal.
		//Save indexing inverse conversion to "actualToIndexedLabels."
		uniqueClassLabelsLocal.resize(seen.size());
		unsigned int i = 0;
		int tempLast = -1;
		for (map<int,int>::iterator itr = seen.begin();
			itr != seen.end(); ++itr)
		{
			uniqueClassLabelsLocal[i] = itr->first;
			actualToIndexedLabels[itr->first] = i;
			++i;
			assert(itr->first > tempLast);	// Should be sorted by iterating over seen.
			tempLast = itr->first;
		}
		return true;
	}

	/* Set samplesByLabel: map from class labels to vectors of the sample indices from that class.
	 */
	bool LocalSDAModel::setSamplesByLabel(map<int,std::vector<unsigned int> > &samplesByLabel,
										  const vector<int> orig_labels)
	{
		int nSamples = (int) orig_labels.size();
		if ( uniqueClassLabelsLocal.size() == 0 )
		{
			return false;
		}
		// Fill samplesByLabel data structure.
		unsigned int sampleIdx = 0;
		for(vector<int>::const_iterator itrI = orig_labels.begin();
			itrI != orig_labels.end(); ++itrI) 
		{
			if(samplesByLabel.count(*itrI))
			{   
				samplesByLabel[*itrI].push_back(sampleIdx);
			} 
			else
			{	
				vector<unsigned int> tmpSampleIdxVect;
				tmpSampleIdxVect.push_back(sampleIdx);
				samplesByLabel[*itrI] = tmpSampleIdxVect;
			}
			++sampleIdx;
		}
		return true;
	}
	
	/* Set centroidIdx: a map from the class label to the index of the local class centroid.
	*/
	bool LocalSDAModel::setClassCentroids(const RealMatrix &simMat, map<int,unsigned int> &centroidIdx)
	{
		if (samplesByLabel.size() == 0)
		{
			return false;
		}
		// Designate a centroid idx for each class
		for(map<int,vector<unsigned int> >::iterator itrL = samplesByLabel.begin();
			itrL != samplesByLabel.end(); ++itrL) 
		{
			vector<unsigned int> tmpSampleIdx = itrL->second;
			unsigned int tmpIdx = 0;
			double tmpMaxVal = (-1)*INF;
			double tmpVal = 0;
			// Find centroid for label L
			for(vector<unsigned int>::const_iterator itrS = tmpSampleIdx.begin();
				itrS != tmpSampleIdx.end(); ++itrS)
			{
				
				tmpVal = sumSims(simMat,tmpSampleIdx,*itrS);
				if (tmpVal > tmpMaxVal) 
				{
					centroidIdx[itrL->first] = *itrS;
					tmpMaxVal = tmpVal;
				}
			}
		}
		return true;
	}

	/* Set centroidsToTestSim: Save similarity(test,centroids) to 1x|L| matrix, |L| = # unique labels in N(test)
	*/
	bool LocalSDAModel::setCentroidsToTestSim(RealMatrix &centroidsToTestSim, const RealMatrix &testSim)
	{
		if (uniqueClassLabelsLocal.size() == 0)
		{
			return false;
		}
		centroidsToTestSim.resize(1,(int)uniqueClassLabelsLocal.size());
		for(vector<int>::const_iterator itr = uniqueClassLabelsLocal.begin();
			itr != uniqueClassLabelsLocal.end(); ++itr)
		{
			unsigned int labelIndex = actualToIndexedLabels[*itr];
			centroidsToTestSim(0,labelIndex) = testSim(0,centroidIdx[*itr]);
		}
		return true;
	}

	/* Set classPriors: a map from the class label to the local probabilitity of that class.
     */
	bool LocalSDAModel::setPriors(map<int,double> &classPriors,int nSamples)
	{
		if ( uniqueClassLabelsLocal.size() == 0 
			|| samplesByLabel.size() == 0)
		{
			return false;
		}
		// Calculate priors, biasing towards equal priors.
		int nClasses = (int)uniqueClassLabelsLocal.size();
		for(vector<int>::const_iterator itrL = uniqueClassLabelsLocal.begin();
			itrL != uniqueClassLabelsLocal.end(); ++itrL)
		{
			classPriors[*itrL] = (double)(samplesByLabel[*itrL].size() + 1)/(nSamples + nClasses);
		}
		// Check that classPriors sum to 1.
		double sumProb = 0;
		for(map<int,double>::const_iterator itrP = classPriors.begin();
			itrP != classPriors.end(); ++itrP)
		{
			sumProb += itrP->second;
		}
		if (sumProb < .99999 || sumProb > 1.00001)
		{
			cerr << "Error class priors don't sum to 1." << endl;
			return false;
		}
		return true;
	}

	/* Set flags for degenerate/simple test sets.
	 * anyOneClassFlag:
	 *   if any label has only one sample -> classify by nearest centroid
	 * onlyOneClassFlag
	 *   if all samples are from one class -> classify as the one class
	 * oneNeighbor
	 *   if there |sample set| == 1 -> classify as the one neighbor
	 */
	bool LocalSDAModel::setDegenerateFlags(const RealMatrix &simTest,
										   bool &anyOneClassFlag, bool &onlyOneClassFlag, 
										   bool &oneNeighbor, bool &allSimsZeroFlag) 
	{
		if (samplesByLabel.size() == 0)
		{
			return false;
		}

		anyOneClassFlag = false;
		onlyOneClassFlag = false;
		oneNeighbor = false;
		allSimsZeroFlag = true;
		
		// Check if the test point has zero similarity to everything.
		for(unsigned int i_neigh = 0; i_neigh < simTest.getColumnNum(); i_neigh++) 
		{
			if (simTest(i_neigh) != 0)
			{
				allSimsZeroFlag = false;
				break;
			}
		}

		for(map<int,vector<unsigned int> >::iterator itr = samplesByLabel.begin();
			itr != samplesByLabel.end(); ++itr) 
		{
			if (itr->second.size() == 1)
			{
				anyOneClassFlag = true;
				break;
			}
		}

		if(samplesByLabel.size() == 1)
		{
			onlyOneClassFlag = true;
			if ( (samplesByLabel.begin()->second).size() == 1)
				oneNeighbor = true;
		}

		return true;
	}

	////////////////////////
	// UTILITY FUNCTIONS //
	///////////////////////

	// Convert class-label-to-weight map to class-label-to-prob map.
	bool LocalSDAModel::computeProbsFromClassWeights(map<int, double> &classWeights, map<int, double> &classProbs)
	{
		// Sum weights.
		double sumWeights = 0;
		for (map<int,double>::const_iterator itr = classWeights.begin(); 
			itr != classWeights.end(); ++itr)
		{
			double classWeight = itr->second;
			if (classWeight < 0) {
				return false; }
			sumWeights += classWeight;
		}
		// Assign probabilities.
		if (sumWeights == 0) {
			//  All centroids have zero sim to test sample.  Give all local classes equal weight.
			for (unsigned int labelIdx = 0; labelIdx < (unsigned int)uniqueClassLabelsLocal.size(); ++labelIdx)
			{
				if (classWeights.count(uniqueClassLabelsLocal[labelIdx]))
				{ 
					classProbs[uniqueClassLabelsLocal[labelIdx]] = 1.0 / uniqueClassLabelsLocal.size();
				}
			}
		} else {
			// Calculate probabilities by normalizing weights.
			for (unsigned int labelIdx = 0; labelIdx < (unsigned int)uniqueClassLabelsLocal.size(); ++labelIdx)
			{
				if (classWeights.count(uniqueClassLabelsLocal[labelIdx]))
				{ 
					classProbs[uniqueClassLabelsLocal[labelIdx]] = classWeights[uniqueClassLabelsLocal[labelIdx]] / sumWeights;
				}
			}
		}
		// Check that sum of probabilities sum to 1.
		double newSumP = 0;
		for (unsigned int labelIdx = 0; labelIdx < (unsigned int)uniqueClassLabelsLocal.size(); ++labelIdx)
		{
			if (classProbs.count(uniqueClassLabelsLocal[labelIdx]))
			{ 
				newSumP += classProbs[uniqueClassLabelsLocal[labelIdx]];
			}
		}
		if ((newSumP < .99999) || (newSumP > 1.0001)) { 
			return false; 
		}
		return true;
	}

	/* Calculates and returns the sum similarity of the sample idx to sample indices vect in vect. */
	double LocalSDAModel::sumSims(const RealMatrix &simMat, vector<unsigned int> &sampleVectIdx, unsigned int sampleIdx)
	{
		double runningSum = 0;
		for(vector<unsigned int>::iterator itrS = sampleVectIdx.begin();
			itrS != sampleVectIdx.end(); ++itrS) 
		{
			runningSum += simMat(sampleIdx,*itrS); //vect[vectIdx]);
		}
		return runningSum;
	}

	
	//////////////////////////////
	// CLASSIFICATION FUNCTIONS //
	//////////////////////////////

	int LocalSDAModel::classify(RealMatrix &simMat, const vector<int> &orig_labels, const RealMatrix &simTest, const RealMatrix &domainVector)
	{
		// For when we aren't interested in probabilities.
		map<int,double> classProbs;	// Just a placeholder.
		int predictedClass = classify(simMat, orig_labels,
								   simTest, domainVector,
								   classProbs);
		return predictedClass;
	}

	// Classify AND get probabilities.
	int LocalSDAModel::classify(RealMatrix &simMat, const vector<int> &orig_labels, const RealMatrix &simTest, const RealMatrix &domainVector, map<int,double> &classProbs)
	{
		int predictedClass = 0;			// This will be the predicted class.
		map<int, double> classWeights;	

		// Take care of special cases first.
		if (oneNeighbor || onlyOneClassFlag)
		{
			// Classify as the only class label in the test's neighborhood.
			std::map<int, unsigned int>::const_iterator itr = centroidIdx.begin();
			predictedClass = itr->first;
			classProbs[predictedClass] = 1;
			return predictedClass;
		}
		else if (anyOneClassFlag || allSimsZeroFlag) 
		{ 
			// Can't compute model; use backup classifier.
			int predictedClass;
			if (backupClassify(simMat, orig_labels, simTest,
				domainVector, predictedClass, classProbs)) {
					return predictedClass;
			} else { 
				cerr << "Error: Backup classifier failure." << endl;
				assert(false);
			}
		} 

		// Construct SDA model
		computeModelParams(simMat, orig_labels, simTest, domainVector);
		double maxPosterior = 0;
		int chosenClassFromModel = -1;
		for (vector<int>::const_iterator itrL = uniqueClassLabelsLocal.begin();
			itrL != uniqueClassLabelsLocal.end(); ++itrL)
		{
			double posterior = posteriorFxn(simTest, domainVector,*itrL);  // Posterior = Prior * Likelihood.
			classWeights[*itrL] = posterior;	// These will be normalized into probabilities.
			if (posterior > maxPosterior)		// Ties are broken by choosing the first.
			{
				maxPosterior = posterior;
				predictedClass = *itrL;
			}
		}
		if (!computeProbsFromClassWeights(classWeights, classProbs)) {
			cerr << "Error: LSDA classifier failure." << endl;
			assert(false);
		}
		// Check SDA results.
		if (maxPosterior == 0) {
			// SDA model has failed (best posterior is 0); call backup classifier.
			if (!backupClassify(simMat, orig_labels, simTest,
				domainVector, predictedClass, classProbs))
			{
				cerr << "Error: Backup classifier failure." << endl;
				assert(false);
			}
		} 
		return predictedClass;
	}

	/* Return the log posterior (Prior*Likelihood).
	 * P(Y=G | s) = P(Y=G) * P(s | Y=G) */
	double LocalSDAModel::posteriorFxn(const RealMatrix &simTest, const RealMatrix &simDomain, int predictedClass) 
	{
		double likelihood;
		likelihood = getNBLikelihoodGivenClass(simTest, simDomain, predictedClass);
		double posterior = likelihood * classPriors[predictedClass]; 
		return posterior;
	}
	
	/* mean constraint function:
	 * E[s] over distribution P(q,u_h(z) | Y = G) 
	 *   = (1/|Q|) * sum( sim(q,u_h(z)) )
	 * Y = label(z), Q = samples of label G in N(z)
	 * domainVect = Q
	 */
	double meanConstraintFxn(double lambda, const RealMatrix &domainVector, double classMean)
	{
		RealMatrix elambdaVect = lambda*domainVector;
		elambdaVect.fxn(exp);

		double sumELambdaVect = elambdaVect.sum();
		double theErr = 0.0;
		if(sumELambdaVect == 0)
		{
			cerr << "Something weird has happened in the meanConstraintFxn";
			lambda = (-1)*INF; 
			return theErr;
		}
		// normalize the pmf 
		elambdaVect /= sumELambdaVect;

		//and then calc the mean:
		double meanpmf = (elambdaVect.dot(domainVector)).sum();

		// absolute error to empirical mean
		theErr = fabs(meanpmf - classMean); 
		return theErr;
	}


	// Backup classifier for when SDA fails.
	bool LocalSDAModel::backupClassify(RealMatrix &simMat, const vector<int> &orig_labels,
								const RealMatrix &simTest,
								const RealMatrix &domainVector,
								int &predictedClass, map<int,double> &classProbs)
	{
		// Classify based on nearest centroid classifier.	
		// Probabilities based on sim weights to each centroid.
		// PJS: Maybe nearest-centroid isn't the best thing to do.  
		double maxTestCentroidSim = (-1)*INF;
		int numMaxSimIdx = 0;
		std::map<int, double> classWeights;
		for (std::map<int, unsigned int>::const_iterator itr = centroidIdx.begin();
			itr != centroidIdx.end(); ++itr)
		{
			classWeights[itr->first] = simTest(0,itr->second);	// Sim used as weight.
			if ( simTest(0,itr->second) > maxTestCentroidSim )
			{
				maxTestCentroidSim = simTest(0,itr->second);
				predictedClass = itr->first;
				numMaxSimIdx = 1;
			} 
			else if ( simTest(0,itr->second) == maxTestCentroidSim )
			{	// Found a tie.
				++numMaxSimIdx;
			}
		}
		// Check for ties.
		if (numMaxSimIdx > 1)
		{	// Break ties by class priors.
			double maxCentroidPrior = (-1)*INF;
			for (std::map<int, unsigned int>::const_iterator itr = centroidIdx.begin();
				itr != centroidIdx.end(); ++itr)
			{
				// If there are equal priors of centroids that are of max similarity to test, 
				// then the first label will be selected.
				if ( simTest(0,itr->second) == maxTestCentroidSim && classPriors[itr->first] > maxCentroidPrior)
				{
					predictedClass = itr->first;
					maxCentroidPrior = classPriors[itr->first];
				}
			}
		}
		// Normalize classWeights into classProbs.
		if (!computeProbsFromClassWeights(classWeights, classProbs)) {
			return false;
		}
		return true;
	}


	///////////////////////
	// GETTER FUNCTIONS //
	//////////////////////

	/* return laplace smoothed prior probability for class label g */
	double LocalSDAModel::getPrior(int g)
	{
		if (! actualToIndexedLabels.count(g))
		{
			return 0.0;
		}
		else
		{
			return classPriors[g];
		}
	}

	/* return the similarity from the test sample to the centroid in class h */
	double LocalSDAModel::getSimTestToH(const RealMatrix &testSimVec, int h)
	{
		if (! actualToIndexedLabels.count(h))
		{
			cerr << "Error: Unknown class label." << endl;
			return 0.0;
		}
		else
		{
			return testSimVec(0,centroidIdx[h]);
		}
	}

	/* Row matrix of similarities from centroid of class h to samples in class g */
	void LocalSDAModel::getHCenterToGSims(const RealMatrix &simMat, RealMatrix &HCenterToGSims, int h, int g) 
	{
		HCenterToGSims = simMat(centroidIdx[h],samplesByLabel[g]);
	}

	/* Vector of sample indices from the specified class g. */
	void LocalSDAModel::getSamplesByLabel(std::vector<unsigned int> &sampleIndices, int g)
	{
		sampleIndices = samplesByLabel[g];
	}

	void LocalSDAModel::getUniqueLabels(std::vector<int> &uniqueLabels)
	{
		uniqueLabels.resize(uniqueClassLabelsLocal.size());
		int count = 0;
		for(vector<int>::iterator itrL = uniqueClassLabelsLocal.begin();
			itrL != uniqueClassLabelsLocal.end(); ++itrL) 
		{
			uniqueLabels[count] = *itrL;
			count += 1;
		}
		return;
	}


	bool getLambdaGamma(const RealMatrix &domainVector, double classMean, double &lambda, double &gamma) 
	{
		double maxSim = domainVector.max();
		double minSim = domainVector.min();
		if (classMean == maxSim) {
			// Degenerate case where label G samples are of max similarity to the centroid H.
			// This could include simDomain of size 1.
			lambda = INF;
			gamma = 0;
		} else if(classMean == minSim) {
			// Degenerate case where label G samples are of min similarity to the centroid H.
			lambda = (-1)*INF;
			gamma = INF;
		} else {
			// Find optimal lambda for local discrete exponential.
			lambda = brent(&idl::meanConstraintFxn,domainVector,classMean,-100,0,100,.0000001,100);
			double theErr = meanConstraintFxn(lambda, domainVector, classMean);
			if (theErr > .01) {
				cout << "Warning: Error large. Possibly because sim domain is large." << endl; }
			RealMatrix elambdaVect = lambda*domainVector;
			elambdaVect.fxn(exp);
			double sumELambdaVect = elambdaVect.sum();
			if (fabs(sumELambdaVect) < .00000001) {
				// PMF is numerically sensitive.
				lambda = (-1)*INF;
				gamma = INF;
			} else {
				gamma = 1/sumELambdaVect;
			}
		}
		return true;
	}


	// Returns the discrete exponential smoothed pmf.
	RealMatrix getDiscreteExponential(const RealMatrix &rawPMF, const RealMatrix &domainVector) 
	{
		// This pmf is designed to have the same domain and mean as our raw pmf.
		// First check input.  The dimensions are important.
		assert((rawPMF.getColumnNum() == domainVector.getColumnNum()));
		assert(rawPMF.getRowNum() == domainVector.getRowNum());
		assert((rawPMF.getRowNum() == 1) || (rawPMF.getColumnNum() == 1));	// rawPMF and domainVector be vectors.
		if (!(rawPMF.sum() > .9999999 && rawPMF.sum() < 1.0000001)) {
			assert(false); 
		}
		// Calculate the mean similarity.
		RealMatrix domainColumnVec;	// The domainVector, but guarenteed to be a column.
		RealMatrix temp;
		if (domainVector.getColumnNum() == 1) {
			// Our input vectors are column matrices.  Great.
			domainColumnVec = domainVector;
			temp = domainColumnVec.dot(rawPMF);	// The dot function should return a matrix.
		} else {
			// We want column matrices.
			domainColumnVec = domainVector.transpose();
			assert(domainColumnVec.getColumnNum() == 1);
			temp = domainColumnVec.dot(rawPMF.transpose());	// The dot function should return a matrix.
		}
		double meanSim = temp.sum();
		
		// Calculate max entropy distribution with same domain and mean.
		double doubleINF = INF;	// Need to assign it to double for comparisons.
		double lambda, gamma;
		idl::getLambdaGamma(domainColumnVec, meanSim, lambda, gamma);
		RealMatrix newPMF;
		if (lambda >= doubleINF) {
			// Special extreme case.
			newPMF.zeros(domainColumnVec.size()[0], 1);	// domainColumnVec sized N*1.
			newPMF(newPMF.getRowNum()-1, 0) = 1; 
		} else if(lambda <= (-1)*doubleINF) {
			// Other extreme case.
			newPMF.zeros(domainColumnVec.size()[0], 1);
			newPMF(0, 0) = 1; 
		} else {
			// We assume this is numerically stable.
			RealMatrix eLambdaVec = lambda*domainColumnVec;
			eLambdaVec.fxn(exp);
			newPMF = gamma * eLambdaVec;
		}
		assert(newPMF.getColumnNum() == 1);
		// Return PMF in the dimensions of the input.
		if (domainVector.getColumnNum() == 1) {
			return newPMF;
		} else {
			return newPMF.transpose();
		}
	}
	/* Return the naive bayes probability of given test sim vector given test sample is in class g.
	 * Prod_i[P(s(X_i,centroidH_i) | Y = G)]
	 */
	double LocalSDAModel::getNBLikelihoodGivenClass(const RealMatrix &testSimVec, const RealMatrix &domainVector, int g)
	{
		if (! actualToIndexedLabels.count(g))
			return 0.0;		
		unsigned int gIndex = actualToIndexedLabels[g];	

		double doubleINF = INF;	// Need to assign it to double for comparisons.
		double likelihood = 1;
		for (vector<int>::const_iterator itrH = uniqueClassLabelsLocal.begin(); 
			itrH != uniqueClassLabelsLocal.end(); ++itrH)
		{
			if (!actualToIndexedLabels.count(*itrH)) {
				cerr << "Error: Unknown local class label." << endl;
				return 0.0; }
			unsigned int hIndex = actualToIndexedLabels[*itrH];
			double testSim = testSimVec(0,centroidIdx[*itrH]);
			double gamma = gammas(gIndex, hIndex);
			double lambda = lambdas(gIndex, hIndex);
			if (lambda >= doubleINF) {
				// Special exteme case.
				if (testSim == domainVector.max()) {
					// Max sim.
					likelihood *= 1;
				} else {
					likelihood = 0; }
			} else if(lambda <= (-1)*doubleINF) {
				// Other extreme case.
				if (testSim == domainVector.min()) {
					// Min sim.
					likelihood *= 1;
				} else {
					likelihood = 0; }
			} else {
				// We assume this is numerically stable.
				likelihood *= gamma * exp(lambda*testSim);
			}
		}
		return likelihood;
	}

	/* Return the probability P(s(X, centroidH)=observed value | Y = G) */
	double LocalSDAModel::getSimProbGivenClass(RealMatrix &testSimVec, int g, int h, const RealMatrix &domainVector)
	{
		double probability;
		if (actualToIndexedLabels.count(g) && actualToIndexedLabels.count(h))
		{
			unsigned int gIndex = actualToIndexedLabels[g];		
			unsigned int hIndex = actualToIndexedLabels[h];		
			double testSim = testSimVec(0,centroidIdx[h]);	// Pick out sim to centroid of H.
			double gamma = gammas(gIndex, hIndex);
			double lambda = lambdas(gIndex, hIndex);
			double doubleINF = INF;	// Need to assign it to double for comparisons.
			if (lambda >= doubleINF) {
				// Special exteme case.
				if (testSim == domainVector.max()) {
					// Max sim.
					probability = 1;
				} else {
					probability = 0; }
			} else if(lambda <= (-1)*doubleINF) {
				// Other extreme case.
				if (testSim == domainVector.min()) {
					// Min sim.
					probability = 1;
				} else {
					probability = 0; }
			} else {
				// We assume this is numerically stable.
				probability = gamma * exp(lambda*testSim);
			}
		}
		else 
		{
			cerr << "Error: shouldn't be here." << endl;
			assert(false);
			probability = 0.0; // PJS: Is this correct?
		}
		assert( probability >= 0 && probability <= 1);
		return probability;
	}
	
}

