// File: LocalSDATest.cpp
// Purpose: Implementation of the LocalSDATest class
// Remarks: This class is designed as an adapter for the LocalSDA class
//			to enable testing on multiple sets of parameters.
// Authors: Ingrid Liu(ingrid@hkn.eecs.berkeley.edu), Peter Sadowski

#include <cmath>
#include <iostream>
#include <algorithm>
#include "RealMatrix.h"
#include "LocalSDATest.h"

using namespace std;

namespace idl
{
	const string LocalSDATest::_ID = "lSDA"; // Define the class ID

	// Register the class instantiator with the factory
	const bool LocalSDATest::_REGISTERED =
		ClassifierTestFactory::registerCreator(LocalSDATest::_ID, &LocalSDATest::create);

	// Destructor
	LocalSDATest::~LocalSDATest()
	{
		// Remarks: The string "factory trick" will never be printed.
		if (!_REGISTERED) cout << "factory trick";
	}

	// Configure the test for a weighted k-NN classifier
	bool LocalSDATest::configure(ConfigParser &config, int secIdx)
	{
		std::cout << "configuring... local\n";
		
		// Get neighborhood sizes to cross-validate
		if (!config.getUIntVec(secIdx, "NeighborhoodSize", neighborhoodSizes))
		{
			cerr << "Error: Missing or incorrect neighborhood size for lSDA classifier "
				 << secIdx << endl;
			return false;
		}

		// Check the retrieved neighborhood sizes
		if (!checkNeighborhoodSizes())
		{
			cerr << "Error: Incorrect neighborhood size for lSDA classifier "
				 << secIdx << endl;
			return false;
		}

		reset(static_cast<unsigned int>(neighborhoodSizes.size()));

		return true;
	}

	// Test the local SDA classifier on all available parameters
	bool LocalSDATest::test(const SimData &data, ofstream &logFile)
	{
		//std::cout << "Testing... \n";
		// Log the weight mode
		logFile << "Local SDA classifier" << endl
				<< "Neighborhood size\tError rate" << endl;
		unsigned int i = 0; // parameter set count

		vector<int> results(data.getTestSetSize());

		for (vector<unsigned int>::const_iterator itrNeighborhoodSize = neighborhoodSizes.begin();
			 itrNeighborhoodSize != neighborhoodSizes.end(); ++itrNeighborhoodSize)
		{
			lsda.setNeighborhoodSize(*itrNeighborhoodSize);
				if (!lsda.classify(data, results)) return false;

				double errorRate = data.getClassificationError(results);
				
				// Log the error rate
				logFile << *itrNeighborhoodSize << "\t"
						<< (errorRate * 100.0) << "%" << endl;

				errorSum(i) += errorRate;
				errorSumOfSquares(i++) += (errorRate * errorRate);
			//}
		}

		logFile << endl;

		++testCount;
		return true;
	}

	// Check the legitimacy of the neighborhood sizes
	bool LocalSDATest::checkNeighborhoodSizes()
	{
		sort(neighborhoodSizes.begin(), neighborhoodSizes.end());

		return neighborhoodSizes.front() > 0;
	}

	// Clear the results with the best parameters
	void LocalSDATest::resetBestParaResults()
	{
		resultsBestPara.clear();
	}

	// Select the best parameters based on the previous tests
	void LocalSDATest::selectBestPara()
	{
		
		ResultType latestResult;
		latestResult.testError = -1.0; // No test with the best parameters yet

		if (0 == testCount)
		{
			// No cross-validation performed
			latestResult.neighborhoodSize = neighborhoodSizes.front();
			latestResult.cvError = 0.0;
			latestResult.cvErrorStdDev = 0.0;
		}
		else // testCount > 0
		{
			unsigned int i;

			// Get the minimum cross-validation error and its standard deviation
			getMinErrorInfo(i, latestResult.cvError, latestResult.cvErrorStdDev);

			// Get the neighborhood size corresponding to the minimum cross-validation error
			latestResult.neighborhoodSize = neighborhoodSizes[i];

		}

		resultsBestPara.push_back(latestResult);
	}

	// Test the weighted k-NN classifier with the best parameters
	bool LocalSDATest::testBestPara(const SimData &data, ofstream &logFile)
	{
		// Log the weight mode
		logFile << "Local SDA classifier" << endl;


		unsigned int k = (resultsBestPara.back()).neighborhoodSize;
		lsda.setNeighborhoodSize(k);

		vector<int> results(data.getTestSetSize());
		RealMatrix probMatrix;
		if (!lsda.classify(data, results, probMatrix)) return false;

		//std::cout << "size " << probMatrix.getRowNum() << " " << probMatrix.getColumnNum() << "asdf";
		logFile << "Classification probabilities:" << endl;
		for (unsigned int i = 0; i < probMatrix.getRowNum(); ++i)
		{
			for (unsigned int j = 0; j < probMatrix.getColumnNum(); ++j)
			{
				logFile << probMatrix(i,j) << " ";
			}
			logFile << endl;
		}
		logFile << "End of classification probabilities." << endl;
		
		
		double errorRate = data.getClassificationError(results, logFile);
		(resultsBestPara.back()).testError = errorRate;

		// Log the error rate
		logFile << "Neighborhood size\tLambda\tError rate" << endl
				<< k << "\t" << (errorRate * 100.0) << "%" << endl << endl;

		return true;
	}

	// Report the test results of all available parameters
	void LocalSDATest::reportResults(ofstream &file, bool verbose)
	{
		file << "Local SDA classifier" << endl;

		if (verbose)
		{
			// Report the results of all the parameter sets
			file << "Neighborhood size\tError rate\tStandard deviation" << endl;

			unsigned int i = 0;
			for (vector<unsigned int>::const_iterator itrNeighborhoodSize = neighborhoodSizes.begin();
				 itrNeighborhoodSize != neighborhoodSizes.end(); ++itrNeighborhoodSize)
			{
				file << *itrNeighborhoodSize << "\t" 
					 << (getError(i) * 100.0) << "%\t" << (getErrorStdDev(i) * 100.0) << "%" << endl;
				++i;
			}
		}
		else
		{
			// Report the best result only
			file << "Neighborhood size: " << (resultsBestPara.back()).neighborhoodSize << endl
				 << "Error rate: " << ((resultsBestPara.back()).cvError * 100.0) << "%" << endl
				 << "Standard deviation: " << ((resultsBestPara.back()).cvErrorStdDev * 100.0) << "%" << endl;
		}

		file << endl;
	}

	// Report the test results with the best parameters
	void LocalSDATest::reportBestParaResults(ofstream &file, bool verbose)
	{
		file << "Local SDA classifier" << endl;

		if (verbose)
		{
			file << "Neighborhood size: " << (resultsBestPara.back()).neighborhoodSize << endl
				 << "Cross-validation error: " << ((resultsBestPara.back()).cvError * 100.0) << "%" << endl
				 << "Standard deviation: " << ((resultsBestPara.back()).cvErrorStdDev * 100.0) << "%" << endl
				 << "Test error: " << ((resultsBestPara.back()).testError * 100.0) << "%" << endl;
		}
		else
		{
			file << "Neighborhood size: " << (resultsBestPara.back()).neighborhoodSize << endl
				 << "Error rate: " << ((resultsBestPara.back()).testError * 100.0) << "%" << endl;
		}

		file << endl;
	}

	// Summarize the results of all the tests with the best parameters
	void LocalSDATest::reportBestParaSummary(ofstream &file, bool verbose)
	{
		file << "Local SDA classifier" << endl;

		if (verbose)
		{
			file << "Neighborhood size\tError rate" << endl;
		}

		unsigned int n = static_cast<unsigned int>(resultsBestPara.size());
		RealMatrix kVec(1, n);
		RealMatrix errorVec(1, n);

		unsigned int i = 0;
		for (vector<ResultType>::const_iterator itr = resultsBestPara.begin();
			 itr != resultsBestPara.end(); ++itr, ++i)
		{
			kVec(i) = static_cast<double>(itr->neighborhoodSize);
			errorVec(i) = itr->testError;

			if (verbose)
			{
				file << itr->neighborhoodSize << "\t"
					 << (itr->testError * 100.0) << "%" << endl;
			}
		}
		
		file << "Average neighborhood size: " << kVec.mean()
			 << " (" << kVec.std() << ")" << endl
			 << "Average error rate: " << (errorVec.mean() * 100.0)
			 << "% (" << (errorVec.std() * 100.0) << "%)" << endl << endl;
	}
}
