// File: PSVM.cpp
// Purpose: Implementation of the PSVM class
// Remarks: This class is designed for PSVM classification.
// Author: Yihua Chen (yhchen@u.washington.edu)
// Last modified: 6/11/2008 12:32:19 AM

#include <cstdlib>
#include <cstdio>
#include <cstring>
#include <fstream>
#include <iostream>
#include <map>
#include "utilities.h"
#include "PSVM.h"
#ifdef WINDOWS
#include <windows.h>
#else // LINUX
#include <unistd.h>
#include <dirent.h>
#endif // WINDOWS

using namespace std;

namespace idl
{
	// Classify the test samples of the similarity data set
	bool PSVM::classify(const SimData &data, vector<int> &results)
	{
		// Get the number of test samples
		unsigned int numTestSamples = data.getTestSetSize();

		results.resize(numTestSamples);

		vector<int> trainingLabels;
		RealMatrix trainingSimMat;
		RealMatrix testSims;

		if (data.isLOOCV())
		{
			for (unsigned int i = 0; i < numTestSamples; ++i)
			{
				data.getTrainingLabels(trainingLabels, i);
				data.getTrainingSimMat(trainingSimMat, i);

				if (!train(trainingSimMat, trainingLabels)) return false;

				data.getTestSimVec(i, testSims);

				if (!predict(testSims, results[i])) return false;
			}
		}
		else
		{
			data.getTrainingLabels(trainingLabels);
			data.getTrainingSimMat(trainingSimMat);

			if (!train(trainingSimMat, trainingLabels)) return false;

			data.getTestSims(testSims);

			if (!predict(testSims, results)) return false;
		}

		// Clean up the working directory
		if (!_clearDirectory(parameters.workingDirectory))
		{
			cerr << "Error: Cannot clean up the working directory" << endl;
			return false;
		}

		return true;
	}

	// Train the PSVM
	bool PSVM::train(const RealMatrix &trainingSimMat, const vector<int> &labels)
	{
		unsigned int numClasses = collectClassInfo(labels);

		if (numClasses < 2)
		{
			cerr << "Error: Total number of classes less than 2" << endl;
			return false;
		}

		string fileNamePrefix = parameters.workingDirectory;
		fileNamePrefix += "sim-";

		// Train binary classification models (one vs. one for multiclass classification)
		for (vector<BinaryModel>::iterator itr = modelMap.begin(); itr != modelMap.end(); ++itr)
		{
			itr->modelName = fileNamePrefix;
			itr->modelName += to_str(itr->classIdx1);
			itr->modelName += '-';
			itr->modelName += to_str(itr->classIdx2);

			if (!writeTrainingDataFile(trainingSimMat, *itr))
			{
				cerr << "Error: Cannot write training data to a file" << endl;
				return false;
			}

			if (!writeLabelFile(*itr))
			{
				cerr << "Error: Cannot write labels to a file" << endl;
				return false;
			}

			// Form the command line for training:
			// [psvm binary] -trainc [model name] -isrelation -c [parameter C] -e [parameter epsilon] \
			// -model [model name] -verbose 0
			string command = parameters.psvmBinary;
			command += " -trainc ";
			command += itr->modelName;
			command += " -isrelation -c ";
			command += to_str(parameters.C);
			command += " -e ";
			command += to_str(parameters.epsilon);
			command += " -model ";
			command += itr->modelName;
			command += " -verbose 0";

			if (system(command.c_str()))
			{
				cerr << "Error: Cannot train a PSVM" << endl;
				return false;
			}

			// Delete the training data and label files
			if (!_deleteFile(itr->modelName, "data")) return false;
			if (!_deleteFile(itr->modelName, "y")) return false;
		}

		return true;
	}

	// Collect the class information from the training samples
	unsigned int PSVM::collectClassInfo(const vector<int> &labels)
	{
		unsigned int numClasses = 0;
		unsigned int sampleIdx = 0;
		map<int, unsigned int> classTable; // map class label to class index

		classMap.clear(); // classMap: map class index to class members
		labelMap.clear(); // labelMap: map class index to class label

		for (vector<int>::const_iterator itr = labels.begin(); itr != labels.end(); ++itr)
		{
			if(classTable.find(*itr) == classTable.end())
			{
				classTable[*itr] = numClasses++;
				labelMap.push_back(*itr);

				vector<unsigned int> members(1, sampleIdx);
				classMap.push_back(members);
			}
			else
			{
				classMap[classTable[*itr]].push_back(sampleIdx);
			}
			++sampleIdx;
		}

		modelMap.clear();

		for (unsigned int i = 0; i < numClasses - 1; ++i)
		{
			for (unsigned int j = i + 1; j < numClasses; ++j)
			{
				BinaryModel model;
				model.classIdx1 = i;
				model.classIdx2 = j;

				modelMap.push_back(model);
			}
		}

		return numClasses;
	}

	// Write the training data to a data file
	bool PSVM::writeTrainingDataFile(const RealMatrix &simMat, const BinaryModel &modelInfo)
	{
		string dataFileName = modelInfo.modelName;
		dataFileName += ".data"; // Form the data file name

		ofstream dataFile(dataFileName.c_str());
		if (!dataFile)
		{
			cerr << "Error: Cannot open " << dataFileName << endl;
			return false;
		}

		vector<unsigned int> samples = classMap[modelInfo.classIdx1];
		samples.insert(samples.end(), classMap[modelInfo.classIdx2].begin(), classMap[modelInfo.classIdx2].end());

		RealMatrix trainingSimMat = simMat(samples, samples);

		dataFile << trainingSimMat;

		return true;
	}

	// Write the test data to a data file
	bool PSVM::writeTestDataFile(const RealMatrix &testSims, const BinaryModel &modelInfo)
	{
		string dataFileName = modelInfo.modelName;
		dataFileName += "-test.data"; // Form the data file name

		ofstream dataFile(dataFileName.c_str());
		if (!dataFile)
		{
			cerr << "Error: Cannot open " << dataFileName << endl;
			return false;
		}

		vector<unsigned int> subset = classMap[modelInfo.classIdx1];
		subset.insert(subset.end(), classMap[modelInfo.classIdx2].begin(), classMap[modelInfo.classIdx2].end());

		RealMatrix testSimMat = testSims.columns(subset);

		dataFile << testSimMat;

		return true;
	}

	// Write the labels of the training samples to a label file
	bool PSVM::writeLabelFile(const BinaryModel &modelInfo)
	{
		string labelFileName = modelInfo.modelName;
		labelFileName += ".y"; // Form the label file name

		ofstream labelFile(labelFileName.c_str());
		if (!labelFile)
		{
			cerr << "Error: Cannot open " << labelFileName << endl;
			return false;
		}

		unsigned int numClass1 = static_cast<unsigned int>(classMap[modelInfo.classIdx1].size());
		unsigned int numClass2 = static_cast<unsigned int>(classMap[modelInfo.classIdx2].size());

		for (unsigned int i = 0; i < numClass1; ++i)
		{
			labelFile << "1" << endl;
		}

		for (unsigned int j = 0; j < numClass2; ++j)
		{
			labelFile << "-1" << endl;
		}

		return true;
	}

	// Read the results of binary classification from a label file
	bool PSVM::readLabelFile(const string &modelName, vector<bool> &results)
	{
		string labelFileName = modelName;
		labelFileName += "-test.yout"; // Form the label file name

		ifstream labelFile(labelFileName.c_str());
		if (!labelFile)
		{
			cerr << "Error: Cannot open " << labelFileName << endl;
			return false;
		}

		unsigned int numTestSamples = static_cast<unsigned int>(results.size());

		int y = 0;
		for (unsigned int i = 0; i < numTestSamples; ++i)
		{
			if (labelFile.eof())
			{
				cerr << "Error: Incorrect predicted output file" << endl;
				return false;
			}

			labelFile >> y;

			if (!labelFile)
			{
				cerr << "Error: Cannot convert to binary class label" << endl;
				return false;
			}

			if (1 == y)
			{
				results[i] = true;
			}
			else if (-1 == y)
			{
				results[i] = false;
			}
			else
			{
				cerr << "Error: Invalid binary class label" << endl;
				return false;
			}
		}

		return true;
	}

	// Predict the label of a test sample
	bool PSVM::predict(const RealMatrix &testSimVec, int &predictedLabel)
	{
		vector<int> result(1);

		if (!predict(testSimVec, result)) return false;

		predictedLabel = result.front();

		return true;
	}

	// Predict the labels of the test samples
	bool PSVM::predict(const RealMatrix &testSims, vector<int> &predictedLabels)
	{
		UIntMatrix votes(testSims.getRowNum(), static_cast<unsigned int>(labelMap.size()), 0);
		vector<bool> results(testSims.getRowNum());

		// Apply binary classification models (one vs. one for multiclass classification)
		for (vector<BinaryModel>::const_iterator itrModel = modelMap.begin();
			 itrModel != modelMap.end(); ++itrModel)
		{
			if (!writeTestDataFile(testSims, *itrModel))
			{
				cerr << "Error: Cannot write test data to a file" << endl;
				return false;
			}

			// Form the command line for predicting:
			// [psvm binary] -predictc [test data file] -model [model name]
			string command = parameters.psvmBinary;
			command += " -predictc ";
			command += itrModel->modelName;
			command += "-test -model ";
			command += itrModel->modelName;
			command += " -verbose 0";

			if (system(command.c_str()))
			{
				cerr << "Error: Cannot test a PSVM" << endl;
				return false;
			}

			if (!readLabelFile(itrModel->modelName, results))
			{
				cerr << "Error: Cannot read binary classification results" << endl;
				return false;
			}

			unsigned int i = 0;
			for (vector<bool>::const_iterator itr = results.begin(); itr != results.end(); ++itr, ++i)
			{
				if (*itr) // Class 1
				{
					++votes(i, itrModel->classIdx1);
				}
				else // Class 2
				{
					++votes(i, itrModel->classIdx2);
				}
			}

			// Delete the test data and label files
			string fileName = itrModel->modelName + "-test";
			if (!_deleteFile(fileName, "data")) return false;
			if (!_deleteFile(fileName, "yout")) return false;
		}

		vote(votes, predictedLabels); // Final voting

		return true;
	}

	// Vote according to the one vs. one classification results
	void PSVM::vote(const UIntMatrix &votes, vector<int> &results)
	{
		unsigned int numTestSamples = votes.getRowNum();
		unsigned int numClasses = votes.getColumnNum();

		for (unsigned int i = 0; i < numTestSamples; ++i)
		{
			unsigned int maxVote = votes(i, 0);
			unsigned int maxPos = 0;

			for (unsigned int j = 1; j < numClasses; ++j)
			{
				if (votes(i, j) > maxVote)
				{
					maxVote = votes(i, j);
					maxPos = j;
				}
			}

#ifdef ENABLE_WARNING
			for (unsigned int k = numClasses - 1; k > maxPos; --k)
			{
				if (votes(i, k) == maxVote)
				{
					cerr << "Warning: Tie in one vs. one voting for PSVM" << endl;
					break;
				}
			}
#endif // ENABLE_WARNING

			results[i] = labelMap[maxPos];
		}
	}

	// Delete a file
	bool PSVM::_deleteFile(const string &fileName, const string &fileExt)
	{
		string fullName = fileName + '.' + fileExt;

		if (0 != remove(fullName.c_str()))
		{
			cerr << "Error: Cannot delete " << fullName << endl;
			return false;
		}

		return true;
	}

	// Clear a directory
	bool PSVM::_clearDirectory(const string &directoryName)
	{
#ifdef WINDOWS
		string dirSpec = directoryName + '*';
		WIN32_FIND_DATA findFileData;
		HANDLE hFind = INVALID_HANDLE_VALUE;

		hFind = FindFirstFile(dirSpec.c_str(), &findFileData);
		if (INVALID_HANDLE_VALUE == hFind)
		{
			cerr << "Error: Invalid file handle (error code: "
				 << GetLastError() << ")" << endl;
			return false;
		}
		else
		{
			do
			{
				if (0 != strcmp(findFileData.cFileName, ".") &&
					0 != strcmp(findFileData.cFileName, ".."))
				{
					string filePath = directoryName + findFileData.cFileName;
					if (0 == DeleteFile(filePath.c_str()))
					{
						cerr << "Error: Cannot delete " << filePath << endl;
						return false;
					}
				}
			}
			while (0 != FindNextFile(hFind, &findFileData));
			FindClose(hFind);
		}
#else // LINUX
		DIR *dirStream;
		struct dirent *dirEntry;

		dirStream = opendir(directoryName.c_str());
		if (NULL == dirStream)
		{
			cerr << "Error: Cannot open directory " << directoryName << endl;
			return false;
		}
		else
		{
			while (dirEntry = readdir(dirStream))
			{
				if (0 != strcmp(dirEntry->d_name, ".") &&
					0 != strcmp(dirEntry->d_name, ".."))
				{
					string filePath = directoryName + dirEntry->d_name;
					if (0 != unlink(filePath.c_str()))
					{
						cerr << "Error: Cannot delete " << filePath << endl;
						return false;
					}
				}
			}
		}
#endif // WINDOWS

		return true;
	}
}
