// File: PSVMTest.h
// Purpose: Definition of the PSVMTest class
// Remarks: This class is designed as an adapter for the PSVM class
//			to enable testing on multiple sets of parameters.
// Author: Yihua Chen (yhchen@u.washington.edu)
// Last modified: 5/27/2008 4:05:00 PM

#ifndef IDL_PSVMTEST_H
#define IDL_PSVMTEST_H

#include <vector>
#include <string>
#include <fstream>
#include "common.h"
#include "ClassifierTest.h"
#include "PSVM.h"

namespace idl
{
	/**
	 * Class for testing an PSVM classifier on similarity data.
	 * An adapter class for the PSVM class to enable testing on multiple
	 * sets of parameters.
	 * @remarks	This class is derived from the ClassifierTest class.
	 * @see		ClassifierTest
	 * @see		ClassifierTestFactory
	 * @see		PSVM
	 */
	class PSVMTest : public ClassifierTest
	{
	public:
		PSVMTest() : ClassifierTest() {}	/**< Constructor. */
		~PSVMTest();						/**< Destructor. */

		/** Clear the results with the best parameters. */
		void resetBestParaResults();

		/** Select the best parameters based on the previous tests. */
		void selectBestPara();

		/**
		 * Configure the test for a PSVM classifier.
		 * @param[in]	config		ConfigParser object containing user configurations
		 * @param[in]	secIdx		section index
		 * @return		true if success; false if failure
		 */
		bool configure(ConfigParser &config, int secIdx);

		/**
		 * Test the PSVM classifier on all available parameters.
		 * @param[in]	data		similarity data to test on
		 * @param[in]	logFile		log file stream
		 * @return		true if success; false if failure
		 */
		bool test(const SimData &data, std::ofstream &logFile);

		/**
		 * Test the PSVM classifier with the best parameters.
		 * @param[in]	data		similarity data to test on
		 * @param[in]	logFile		log file stream
		 * @return		true if success; false if failure
		 */
		bool testBestPara(const SimData &data, std::ofstream &logFile);

		/**
		 * Report the test results of all available parameters.
		 * @param[in]	file		output file stream
		 * @param[in]	verbose		flag for verbose/concise output: true for verbose;
		 *							false for concise (false by default)
		 * @remarks		For concise output, only the best result and its corresponding
		 *				parameters are reported. For verbose output, the results of all
		 *				available parameter sets are reported.
		 */
		void reportResults(std::ofstream &file, bool verbose = false);

		/**
		 * Report the test results with the best parameters.
		 * @param[in]	file		output file stream
		 * @param[in]	verbose		flag for verbose/concise output: true for verbose;
		 *							false for concise (false by default)
		 */
		void reportBestParaResults(std::ofstream &file, bool verbose = false);

		/**
		 * Summarize the results of all the tests with the best parameters.
		 * @param[in]	file		output file stream
		 * @param[in]	verbose		flag for verbose/concise output: true for verbose;
		 *							false for concise (false by default)
		 */
		void reportBestParaSummary(std::ofstream &file, bool verbose = false);

	private:
		PSVM psvm;						/**< PSVM classifier. */
		std::vector<double> Cs;			/**< Vector of hyperparameter C. */
		std::vector<double> epsilons;	/**< Vector of hyperparameter epsilon. */

		/** Structure type for test results with the best parameters. */
		struct ResultType
		{
			double C;				/**< Hyperparameter C. */
			double epsilon;			/**< Hyperparameter epsilon. */
			double cvError;			/**< Cross-validation error. */
			double cvErrorStdDev;	/**< Standard deviation of the cross-validation error. */
			double testError;		/**< Test error. */
		};

		std::vector<ResultType> resultsBestPara; /**< Test results with the best parameters. */

	private:
		/**
		 * Sort the elements of a real vector and check their positivity.
		 * @param[in, out]	vec		real vector to sort and check
		 * @return			true if all the elements are positive;
		 *					false if at least one element is not positive
		 */
		static bool checkPositive(std::vector<double> &vec);

		/**
		 * Create an object of the PSVMTest class.
		 * @return	pointer to a new object of the PSVMTest class
		 * @remarks	This is the creator function to be registered with the factory.
		 */
		static ClassifierTest *create()
		{
			return new PSVMTest;
		}
		static const std::string _ID;	/**< Class ID ("PSVM") */
		static const bool _REGISTERED;	/**< Flag for creator registration */
	};
}

#endif // IDL_PSVMTEST_H
