// File: RealMatrix.h
// Purpose: Definition of the RealMatrix class
// Author: Yihua Chen (yhchen@u.washington.edu)
// Last modified: 5/24/2008 5:54:15 PM

#ifndef IDL_REALMATRIX_H
#define IDL_REALMATRIX_H

#include "cxcore.h" // OpenCV CXCORE header file
#include "Matrix.h"

namespace idl
{
	/**
	 * Class for real-valued matrix.
	 * @remarks	This class is derived from the Matrix class template.
	 * @see		Matrix
	 */
	class RealMatrix : public Matrix<double>
	{
	public:
		RealMatrix() : Matrix<double>() {} /**< Default constructor. */

		/**
		 * Constructor.
		 * @param[in]	m	row number
		 * @param[in]	n	column number
		 */
		RealMatrix(unsigned int m, unsigned int n)
			: Matrix<double>(m, n)
		{
		}

		/**
		 * Constructor.
		 * @param[in]	m		row number
		 * @param[in]	n		column number
		 * @param[in]	initVal	initial value for all the elements
		 */
		RealMatrix(unsigned int m, unsigned int n, double initVal)
			: Matrix<double>(m, n, initVal)
		{
		}

		/**
		 * Copy constructor.
		 * @param[in]	mat	real-valued matrix to copy from
		 */
		RealMatrix(const RealMatrix &mat) : Matrix<double>(mat)
		{
		}

		/**
		 * Conversion constructor.
		 * @param[in]	mat real-valued matrix of the parent class
		 */
		RealMatrix(const Matrix<double> &mat) : Matrix<double>(mat)
		{
		}

		~RealMatrix() {} /**< Destructor. */

		/**
		 * Assignment operator.
		 * Copy the contents of another real-valued matrix.
		 * @param[in]	mat real-valued matrix to copy from
		 * @return		reference to itself
		 */
		RealMatrix & operator = (const RealMatrix &mat)
		{
			nRows = mat.nRows;
			nColumns = mat.nColumns;
			elements = mat.elements;

			return *this;
		}

		double sum() const;		/**< Get the sum of all the elements. */
		double max() const;		/**< Get the maximum of all the elements. */
		double min() const;		/**< Get the minimum of all the elements. */
		double mean() const;	/**< Get the mean of all the elements. */
		double var() const;		/**< Get the variance of all the elements. */
		double std() const;		/**< Get the standard deviation of all the elements. */

		double sumRow(unsigned int i) const;	/**< Get the sum of a row. */
		double maxRow(unsigned int i) const;	/**< Get the maximum of a row. */
		double minRow(unsigned int i) const;	/**< Get the minimum of a row. */
		double meanRow(unsigned int i) const;	/**< Get the mean of a row. */

		double sumColumn(unsigned int j) const;		/**< Get the sum of a column. */
		double maxColumn(unsigned int j) const;		/**< Get the maximum of a column. */
		double minColumn(unsigned int j) const;		/**< Get the minimum of a column. */
		double meanColumn(unsigned int j) const;	/**< Get the mean of a column. */

		void unique(RealMatrix &uniqueVals) const;

		void fxn(double (*func)(double));

		RealMatrix dot(const RealMatrix &B);

		/**
		 * Get the trace.
		 * @return		trace
		 * @attention	Trace is only defined for square matrix.
		 */
		double trace() const;

		RealMatrix operator - () const; /**< Unary negation. */
		
		//@{
		/** Create an identity matrix. */
		RealMatrix & eye();
		RealMatrix & eye(unsigned int m);
		RealMatrix & eye(unsigned int m, unsigned int n);
		//@}

		//@{
		/** Create a matrix of all ones. */
		RealMatrix & ones();
		RealMatrix & ones(unsigned int m);
		RealMatrix & ones(unsigned int m, unsigned int n);
		//@}

		//@{
		/** Create a matrix of all zeros. */
		RealMatrix & zeros();
		RealMatrix & zeros(unsigned int m);
		RealMatrix & zeros(unsigned int m, unsigned int n);
		//@}

		RealMatrix & negate(); /**< Negate all the elements. */
		RealMatrix & operator += (double v);
		RealMatrix & operator -= (double v);
		RealMatrix & operator *= (double v);
		RealMatrix & operator /= (double v);
		RealMatrix & operator += (const RealMatrix &A);
		RealMatrix & operator -= (const RealMatrix &A);
		RealMatrix & operator *= (const RealMatrix &A);

		CvMat *dumpCvMat() const;
		void swallowCvMat(const CvMat *mat);

		void eig(RealMatrix &d) const;
		void eig(RealMatrix &V, RealMatrix &d) const;
	};

	RealMatrix operator + (const RealMatrix &A, double v);
	RealMatrix operator - (const RealMatrix &A, double v);
	RealMatrix operator * (const RealMatrix &A, double v);
	RealMatrix operator / (const RealMatrix &A, double v);

	RealMatrix operator + (double v, const RealMatrix &A);
	RealMatrix operator - (double v, const RealMatrix &A);
	RealMatrix operator * (double v, const RealMatrix &A);

	RealMatrix operator + (const RealMatrix &A, const RealMatrix &B);
	RealMatrix operator - (const RealMatrix &A, const RealMatrix &B);
	RealMatrix operator * (const RealMatrix &A, const RealMatrix &B);

	bool solve(const RealMatrix &A, const RealMatrix &b, RealMatrix &x, bool svd = false);
}

#endif // IDL_REALMATRIX_H
