// File: RegLocalSDA.h
// Purpose: Definition of the RegLocalSDA class
// Remarks: This class is designed for Regularized Local SDA classification.
// Author: Peter Sadowski peterjsadowski@gmail.com

#ifndef IDL_RegLocalSDA_H
#define IDL_RegLocalSDA_H

#include <vector>
#include <fstream>
#include "cxcore.h"	// OpenCV CXCORE header file
#include "ml.h"		// OpenCV MLL header file
#include "RealMatrix.h"
#include "Classifier.h"
#include "SimData.h"
#include "LocalSDAModel.h"

//#include <vector>
//#include <fstream>
//#include "RealMatrix.h"
//#include "Classifier.h"
//#include "SimData.h"
//#include "LocalSDAModel.h"
//#include <string>

namespace idl
{

	/**
	 * Class for the Local SDA classifier.
	 * @remarks	This class is derived from the Classifier class.
	 * @see		Classifier
	 */
	class RegLocalSDA : public Classifier
	{
	public:
		/** Structure type for the Local SDA parameters. */
		struct ParaType
		{
			unsigned int neighborhoodSize;	/**< Neighborhood size. */
			double regParam;				/**< Regularization parameter  */
			bool marginalFlag;
			bool jointFlag;
			bool peterFlag;
		};

		/** Default constructor. */
		RegLocalSDA()
		{
			parameters.neighborhoodSize = 1;
			parameters.regParam = -1.0;
			parameters.marginalFlag = false;
			parameters.jointFlag = false;
			parameters.peterFlag = false;
		}

		/**
		 * Constructor.
		 * @param	para	parameters for the Local SDA classifier
		 */
		RegLocalSDA(const ParaType &para)
		{
			parameters = para;
		}

		~RegLocalSDA() {}	/**< Destructor. */

		/**
		 * Set the neighborhood size.
		 * @param[in]	k	neighborhood size
		 */
		void setNeighborhoodSize(unsigned int k)
		{
			if (k > 0)
			{
				parameters.neighborhoodSize = k;
			}
		}

		/**
		 * Set the regularization parameter.
		 * @param[in]	regParam	regularization parameter regParam for l-SDA
		 */
		void setRegPara(double regParam)
		{
			parameters.regParam = regParam;
		}

		void setFlag(bool marginalFlag, bool jointFlag, bool peterFlag)
		{
			parameters.marginalFlag = marginalFlag;
			parameters.jointFlag = jointFlag;
			parameters.peterFlag = peterFlag;
		}

		/**
		 * Set the parameters for the Local SDA classifier.
		 * @param[in]	para	parameters for the Local SDA classifier
		 */
		void setPara(const ParaType &para)
		{
			parameters = para;
		}
		
		// Configure the regularization.
		bool configure(const SimData &data);

		/**
		 * Classify the test samples of the similarity data set.
		 * @param[in]	data		Similarity data set.
		 * @param[out]	results		Vector to store the estimated class labels.
		 * @param[out]  classProbMatrix  nTestSamples*nUniqueLabels matrix of probabilities.
		 * @return		true if success; false if failure
		 */
		bool classify(const SimData &data, std::vector<int> &results);
		bool classify(const SimData &data, std::vector<int> &results, RealMatrix &classProbMatrix);

	private:
		/**
		 * Check if the given labels belong to one class.
		 * @param[in]	labels		vector of the given class labels
		 * @return		true if the given labels belong to one class;
		 *				false if they belong to more than one class
		 */
		static bool isOneClass(const std::vector<int> &labels);

		double regMarginalProbability(LocalSDAModel &lsda, int g, RealMatrix &testSimVec, const RealMatrix &simDomain);
		double regJointProbability(LocalSDAModel &lsdaint, int g, RealMatrix &testSimVec, const RealMatrix &simDomain);
		void regMarginalJointProbability(LocalSDAModel &lsda, int g, RealMatrix &testSimVec, const RealMatrix &simDomain,
										 std::vector<double> &probabilities);
		double regPeterProbability(LocalSDAModel &lsda, int g, RealMatrix &testSimVec, const RealMatrix &simDomain);
 
		// Class variables.
		ParaType parameters;
		// Average of locally smoothed sim value PMFs over all training neighborhoods.
		std::map<std::vector<int>,idl::RealMatrix> PAvg_ICML09; 
		// Bin the sim values from all training neighborhoods, then fit exponential.
		std::map<std::vector<int>,idl::RealMatrix> PAvg_Peter;	
		std::map<int,double> regLabelPriors;  // "Smoothed" prior over entire training set.
		std::map<double,unsigned int> simIdx;
		LocalSDAModel rlsdaModel;	/**< The model for this test sample. */
	};
}

#endif // IDL_RegLocalSDA_H
