// File: RegLocalSDABN.cpp
// Purpose: Implementation of the RegLocalSDABN class
// Remarks: This class is designed for RegLocalSDABN classification.
// Author: Ingrid (ingrid@hkn.eecs.berkeley.edu)
// Last modified: 3/6/2008 4:52:21 PM

#include <cstdlib>
#include <iostream>
#include <algorithm>
#include "RealMatrix.h"
#include "RegLocalSDABN.h"
#include "LocalSDAModel.h"
#define INF 1.7E308;
#define REGDEFAULT 1e-3;

using namespace std;

namespace idl
{

	bool RegLocalSDABN::classify(const SimData &data, vector<int> &results, bool jointFlag, RealMatrix &probMatrix)
	{
		// Check the neighborhood size: the neighborhood size must be
		// less than or equal to the training set size
		if (data.getTrainingSetSize() < parameters.neighborhoodSizes.back())
		{
			cerr << "Error: Neighborhood size too large" << endl;
			return false;
		}

		// Get the number of test samples
		unsigned int numTestSamples = data.getTestSetSize();
		results.resize(numTestSamples);

		vector<int> uniqueLabels; 
		data.getUniqueLabels(uniqueLabels);
		unsigned int nLabels = (unsigned int)uniqueLabels.size();

		RegLocalSDA::setRegPara(parameters.regParam);

		vector<int> dummyResults;
		probMatrix.zeros(numTestSamples, nLabels);
		for (vector<unsigned int>::const_iterator itrN = parameters.neighborhoodSizes.begin();
				itrN != parameters.neighborhoodSizes.end(); ++itrN)
		{
			RegLocalSDA::setNeighborhoodSize(*itrN);
			RegLocalSDA::configure(data);

			RealMatrix tmpP;
			RegLocalSDA::classify(data,dummyResults,tmpP);
			
			probMatrix += tmpP;			
		}
		for (unsigned int i = 0; i < probMatrix.getRowNum(); ++i)
		{
			double sumRow = probMatrix.sumRow(i);
			for (unsigned int j = 0; j < probMatrix.getColumnNum(); ++j)
			{
				if (probMatrix(i,j) != 0)
					probMatrix(i,j) = probMatrix(i,j) / sumRow;
			}
		}
		
		for (unsigned int i = 0; i < numTestSamples; ++i)
		{
			results[i] = maxLabel(uniqueLabels,probMatrix.row(i));
		}
		return true;
	}

	bool RegLocalSDABN::classify(const SimData &data, vector<int> &results, bool jointFlag)
	{
		RealMatrix probMatrix;
		return classify(data,results,jointFlag,probMatrix);
	}

	//bool RegLocalSDABN::classify(const SimData &data, vector<int> &results, bool jointFlag)
	//{
	//	// Check the neighborhood size: the neighborhood size must be
	//	// less than or equal to the training set size
	//	if (data.getTrainingSetSize() < parameters.neighborhoodSizes.back())
	//	{
	//		cerr << "Error: Neighborhood size too large" << endl;
	//		return false;
	//	}

	//	// Get the number of test samples
	//	unsigned int numTestSamples = data.getTestSetSize();
	//	results.resize(numTestSamples);

	//	vector<int> uniqueLabels; 
	//	data.getUniqueLabels(uniqueLabels);
	//	unsigned int nLabels = (unsigned int)uniqueLabels.size();

	//	RealMatrix probabilities;
	//	map<unsigned int, RealMatrix> testSampleProbabilities;
	//	for (unsigned int i = 0; i < numTestSamples; ++i)
	//	{
	//		RealMatrix tmpP(1,nLabels);
	//		testSampleProbabilities[i] = tmpP; 
	//	}

	//	RegLocalSDA::setRegPara(parameters.regParam);
	//	for (vector<unsigned int>::const_iterator itrN = parameters.neighborhoodSizes.begin();
	//			itrN != parameters.neighborhoodSizes.end(); ++itrN)
	//	{
	//		RegLocalSDA::setNeighborhoodSize(*itrN);
	//		RegLocalSDA::configure(data);
	//		RealMatrix tmpP;
	//		for (unsigned int i = 0; i < numTestSamples; ++i)
	//		{
	//			RegLocalSDA::getProbabilities(data,i,jointFlag,uniqueLabels,tmpP);
	//			testSampleProbabilities[i] += tmpP;
	//		}			
	//	}

	//	
	//	for (unsigned int i = 0; i < numTestSamples; ++i)
	//	{
	//		results[i] = maxLabel(uniqueLabels,testSampleProbabilities[i]);
	//	}
	//	return true;
	//}

	// Return the label with the largest probability
	int RegLocalSDABN::maxLabel(const vector<int> &labels, const RealMatrix &probabilities)
	{
		double tmpMaxP = -DBL_MAX;
		int tmpMaxLabel = -1;
		unsigned int tmpIdx = 0;

		for (unsigned int tmpIdx = 0; tmpIdx < probabilities.size()[1]; ++tmpIdx)
		{
			if (probabilities(0,tmpIdx) > tmpMaxP)
			{
				tmpMaxP = probabilities(0,tmpIdx);
				tmpMaxLabel = tmpIdx;
			}
		}
		return labels[tmpMaxLabel];
	}	

	// dummy class: should never be called 
	bool RegLocalSDABN::classify(const SimData &data, vector<int> &results)
	{
		return false;
	}


}
