// File: RegLocalSDABNTest.h
// Purpose: Definition of the RegLocalSDABNTest class
// Remarks: This class is designed as an adapter for the LocalSDA class
//			to enable testing on multiple sets of parameters.
// Author: Yihua Chen (yhchen@u.washington.edu)
// Last modified: 5/27/2008 4:24:22 PM

#ifndef IDL_REGLOCALSDABNTEST_H
#define IDL_REGLOCALSDABNTEST_H

#include <vector>
#include <string>
#include <fstream>
#include "ClassifierTest.h"
#include "RegLocalSDABN.h"
#include <string>

namespace idl
{
	/**
	 * Class for testing a weighted k-NN classifier on similarity data.
	 * An adapter class for the LocalSDA class to enable testing on multiple
	 * sets of parameters.
	 * @remarks	This class is derived from the ClassifierTest class.
	 * @see		ClassifierTest
	 * @see		ClassifierTestFactory
	 * @see		LocalSDA
	 */
	class RegLocalSDABNTest : public ClassifierTest
	{
	public:
		RegLocalSDABNTest() : ClassifierTest() {}	/**< Constructor. */
		~RegLocalSDABNTest();						/**< Destructor. */

		/** Clear the results with the best parameters. */
		void resetBestParaResults();

		/** Select the best parameters based on the previous tests. */
		void selectBestPara();

		/**
		 * Configure the test for a weighted k-NN classifier.
		 * @param[in]	config		ConfigParser object containing user configurations
		 * @param[in]	secIdx		section index
		 * @return		true if success; false if failure
		 */
		bool configure(ConfigParser &config, int secIdx);

		/**
		 * Test the weighted k-NN classifier on all available parameters.
		 * @param[in]	data		similarity data to test on
		 * @param[in]	logFile		log file stream
		 * @return		true if success; false if failure
		 */
		bool test(const SimData &data, std::ofstream &logFile);

		/**
		 * Test the weighted k-NN classifier with the best parameters.
		 * @param[in]	data		similarity data to test on
		 * @param[in]	logFile		log file stream
		 * @return		true if success; false if failure
		 */
		bool testBestPara(const SimData &data, std::ofstream &logFile);

		/**
		 * Report the test results of all available parameters.
		 * @param[in]	file		output file stream
		 * @param[in]	verbose		flag for verbose/concise output: true for verbose;
		 *							false for concise (false by default)
		 * @remarks		For concise output, only the best result and its corresponding
		 *				parameters are reported. For verbose output, the results of all
		 *				available parameter sets are reported.
		 */
		void reportResults(std::ofstream &file, bool verbose = false);

		/**
		 * Report the test results with the best parameters.
		 * @param[in]	file		output file stream
		 * @param[in]	verbose		flag for verbose/concise output: true for verbose;
		 *							false for concise (false by default)
		 */
		void reportBestParaResults(std::ofstream &file, bool verbose = false);

		/**
		 * Summarize the results of all the tests with the best parameters.
		 * @param[in]	file		output file stream
		 * @param[in]	verbose		flag for verbose/concise output: true for verbose;
		 *							false for concise (false by default)
		 */
		void reportBestParaSummary(std::ofstream &file, bool verbose = false);

	private:
		/**
		 * Check the legitimacy of the regularization parameters.
		 * @return	true if all the regularization parameters are within the legal range;
		 *			false if at least one regularization parameter is out of the legal range
		 */
		//bool checkLambdas();

		/**
		 * Check the legitimacy of the neighborhood sizes.
		 * @return	true if all the neighborhood sizes are within the legal range;
		 *			false if at least one neighborhood size is out of the legal range
		 */
		bool checkNeighborhoodSizes();
		bool setRegularizationFlags(std::string regMode, bool &regularizeFlag, bool &marginalFlag, bool &jointFlag, bool &bothFlag);

	private:
		RegLocalSDABN rlsda; /**< Regularized Local SDA with Bayesian Neighborhoods classifier. */
		std::vector<double> regParams; /**< Vector of regularization parameters. */
		std::vector<unsigned int> neighborhoodSizes; /**< Vector of neighborhood sizes. */
		bool regularizeFlag;
		bool marginalFlag;
		bool jointFlag;
		bool bothFlag;

		/** Structure type for test results with the best parameters. */
		struct ResultType
		{
			//unsigned int neighborhoodSizes;	/**< Neighborhood size. */
			double regParam;					/**< Regularization parameter. */
			double cvError;					/**< Cross-validation error. */
			double cvErrorStdDev;			/**< Standard deviation of the cross-validation error. */
			double testError;				/**< Test error. */
		};

		std::vector<ResultType> resultsBestPara; /**< Test results with the best parameters. */

	private:
		/**
		 * Create an object of the RegLocalSDABNTest class.
		 * @return	pointer to a new object of the RegLocalSDABNTest class
		 * @remarks	This is the creator function to be registered with the factory.
		 */
		static ClassifierTest *create()
		{
			return new RegLocalSDABNTest;
		}
		static const std::string _ID;	/**< Class ID ("kNN") */
		static const bool _REGISTERED;	/**< Flag for creator registration */
	};
}

#endif // IDL_LOCALSDATEST_H
