// File: SVM.h
// Purpose: Definition of the SVM class
// Remarks: This class is designed for SVM classification.
// Author: Yihua Chen (yhchen@u.washington.edu)
// Last modified: 3/6/2008 12:04:57 AM

#ifndef IDL_SVM_H
#define IDL_SVM_H

#include <cstring>
#include <vector>
#include "cxcore.h"	// OpenCV CXCORE header file
#include "ml.h"		// OpenCV MLL header file
#include "RealMatrix.h"
#include "Classifier.h"
#include "SimData.h"
#include "SimEmbed.h"

namespace idl
{
	/**
	 * Class for the SVM classifier.
	 * @remarks	This class is derived from the Classifier class.
	 * @see		Classifier
	 */
	class SVM : public Classifier
	{
	public:
		/**
		 * Enumeration type for kernels:
		 * - Linear kernel: k(x, x') = <x, x'>
		 * - Polynomial kernel: k(x, x') = (gamma * <x, x'> + r)^d
		 * - Gaussian radial basis function kernel: k(x, x') = exp(-gamma * ||x - x'||^2)
		 * - Sigmoid kernel: k(x, x') = tanh(gamma * <x, x'> + r)
		 * - Similarity kernel with eigenvalue shift
		 * - Similarity kernel with eigenvalue clip
		 * - Similarity kernel with eigenvalue flip
		 * @remarks	All the kernel parameters are in the structure ParaType.
		 */
		enum KernelType
		{
			KERNEL_LINEAR,		/**< linear kernel */
			KERNEL_POLY,		/**< polynomial kernel */
			KERNEL_RBF,			/**< Gaussian radial basis function kernel */
			KERNEL_SIGMOID,		/**< sigmoid kernel */
			KERNEL_SIM_SHIFT,	/**< similarity kernel with eigenvalue shift */
			KERNEL_SIM_CLIP,	/**< similarity kernel with eigenvalue clip */
			KERNEL_SIM_FLIP		/**< similarity kernel with eigenvalue flip */
		};

		/** Structure type for SVM parameters. */
		struct ParaType
		{
			KernelType kernel;					/**< Kernel type. */
			double gamma;						/**< Kernel parameter for polynomial, RBF or sigmoid kernel. */
			double r;							/**< Kernel parameter for polynomial or sigmoid kernel. */
			double d;							/**< Kernel parameter for polynomial kernel. */
			double C;							/**< Penalty parameter for C-SVM. */
			bool isStrictlySupervised;			/**< Flag for strictly supervised learning. */
			bool usingCompensatedProjection;	/**< Flag for compensated projection. */
		};

		/** Default constructor. */
		SVM()
		{
			parameters.kernel = KERNEL_LINEAR;
			parameters.gamma = 1.0;
			parameters.r = 1.0;
			parameters.d = 1.0;
			parameters.C = 1.0;
			parameters.isStrictlySupervised = false;
			parameters.usingCompensatedProjection = false;
		}

		/**
		 * Constructor.
		 * @param	para	parameters for the SVM classifier
		 */
		SVM(const ParaType &para)
		{
			parameters = para;
			_setEmbedder();
		}

		~SVM() {}	/**< Destructor. */

		/**
		 * Set the flag for strictly supervised learning.
		 * @param[in]	flag	flag for strictly supervised learning
		 */
		void setStrictlySupervisedFlag(bool flag)
		{
			parameters.isStrictlySupervised = flag;
		}

		/**
		 * Set the flag for compensated projection.
		 * @param[in]	flag	flag for compensated projection
		 */
		void setCompensatedProjectionFlag(bool flag)
		{
			parameters.usingCompensatedProjection = flag;
			embedder.setCompensationFlag(flag);
		}

		/**
		 * Set the type of the kernel.
		 * @param[in]	t	type of the kernel
		 */
		void setKernelType(KernelType t)
		{
			parameters.kernel = t;
			_setEmbedder();
		}

		/**
		 * Set one of the floating-point parameters.
		 * @param[in]	name	name of the floating-point parameter
		 * @param[in]	value	value of the floating-point parameter
		 */
		void setPara(const char *name, double value)
		{
			if (0 == ::strcmp(name, "C"))
			{
				parameters.C = value;
			}
			else if (0 == ::strcmp(name, "gamma"))
			{
				parameters.gamma = value;
			}
			else if (0 == ::strcmp(name, "r"))
			{
				parameters.r = value;
			}
			else if (0 == ::strcmp(name, "d"))
			{
				parameters.d = value;
			}
		}

		/**
		 * Set the parameters for the SVM classifier.
		 * @param[in]	para	parameters for the SVM classifier
		 */
		void setPara(const ParaType &para)
		{
			parameters = para;
			_setEmbedder();
		}

		/**
		 * Get the type of the kernel.
		 * @return	type of the kernel
		 */
		KernelType getKernelType() const
		{
			return parameters.kernel;
		}

		/**
		 * Get the name of the kernel.
		 * @return	string representation of the kernel
		 */
		const char *getKernelName() const
		{
			return kernelNames[parameters.kernel];
		}

		/**
		 * Check if the SVM classifier is strictly supervised.
		 * @return		true if the SVM classifier is strictly supervised; false if not strictly supervised
		 */
		bool isStrictlySupervised() const
		{
			return parameters.isStrictlySupervised;
		}

		/**
		 * Check if the compensated projection is used.
		 * @rerturn		true if the compensated projection is used; false if not used
		 */
		bool usingCompensatedProjection() const
		{
			return parameters.usingCompensatedProjection;
		}

		/**
		 * Classify the test samples of the similarity data set.
		 * @param[in]	data		similarity data set
		 * @param[out]	results		vector to store the estimated class labels
		 * @return		true if success; false if failure
		 */
		bool classify(const SimData &data, std::vector<int> &results);

	private:
		/**
		 * Train the SVM.
		 * @param[in]	trainingSamples		matrix of the training samples (each row is a training sample)
		 * @param[in]	labels				labels of the training samples
		 * @return		true if success; false if failure
		 */
		bool train(const RealMatrix &trainingSamples, const std::vector<int> &labels);

		/**
		 * Predict the label of a test sample.
		 * @param[in]	testSample		test sample
		 * @return		label predicted for the given test sample
		 */
		int predict(const RealMatrix &testSample);

		/** Set the embedding method of the embedder. */
		void _setEmbedder()
		{
			if (KERNEL_SIM_SHIFT == parameters.kernel)
			{
				embedder.setSpectrumTransformMode(SimEmbed::SPECTRUM_SHIFT);
			}
			else if (KERNEL_SIM_CLIP == parameters.kernel)
			{
				embedder.setSpectrumTransformMode(SimEmbed::SPECTRUM_CLIP);
			}
			else if (KERNEL_SIM_FLIP == parameters.kernel)
			{
				embedder.setSpectrumTransformMode(SimEmbed::SPECTRUM_FLIP);
			}
			embedder.setCompensationFlag(parameters.usingCompensatedProjection);
		}
		
	private:
		static const char *kernelNames[]; /** String representations of the kernels. */

		ParaType parameters;	/**< Parameters for the SVM classifier. */
		SimEmbed embedder;		/**< Embedder for embedding similarity data. */
		CvSVM svm;				/**< Multi-class SVM from OpenCV MLL. */
	};
}

#endif // IDL_SVM_H
