// File: SVMKNN.cpp
// Purpose: Implementation of the SVMKNN class
// Remarks: This class is designed for SVM-KNN classification.
// Author: Yihua Chen (yhchen@u.washington.edu)
// Last modified: 3/6/2008 4:52:21 PM

#include <cstdlib>
#include <iostream>
#include <algorithm>
#include "SVMKNN.h"

using namespace std;

namespace idl
{
	// Initialize the string representations of the embedding methods
	const char *SVMKNN::embeddingNames[] = {
		"eigenvalue shift",
		"eigenvalue clip",
		"eigenvalue flip",
		"no embedding" };

	// Classify the test samples of the similarity data set
	bool SVMKNN::classify(const SimData &data, vector<int> &results)
	{
		// Check the neighborhood size: the neighborhood size must be
		// less than or equal to the training set size
		if (data.getTrainingSetSize() < parameters.neighborhoodSize)
		{
			cerr << "Error: Neighborhood size too large" << endl;
			return false;
		}

		// Get the number of test samples
		unsigned int numTestSamples = data.getTestSetSize();

		results.resize(numTestSamples);

		if (1 == parameters.neighborhoodSize)
		{
			// 1-NN
			for (unsigned int i = 0; i < numTestSamples; ++i)
			{
				results[i] = data.get1NNLabel(i);
			}
		}
		else
		{
			RealMatrix simMat(parameters.neighborhoodSize + 1,
							  parameters.neighborhoodSize + 1);
			vector<int> labels(parameters.neighborhoodSize);

			for (unsigned int i = 0; i < numTestSamples; ++i)
			{
				data.getNNLabels(i, parameters.neighborhoodSize, labels);

				if (isOneClass(labels))
				{
					// All the neighbors belong to one class
					results[i] = labels.front();
				}
				else
				{
					data.getNNTestSimMat(i, parameters.neighborhoodSize, simMat);
					results[i] = estimateClassLabel(simMat, labels);
				}
			}
		}

		return true;
	}

	// Estimate the class label of a test sample
	int SVMKNN::estimateClassLabel(const RealMatrix &simMat, const vector<int> &labels)
	{
		RealMatrix embeddedSamples;

		if (EMBED_NONE == parameters.embeddingMode)
		{
			// Treat the similarity vectors directly as the feature vectors
			embeddedSamples = simMat.columns(0, simMat.getColumnNum() - 2);
		}
		else
		{
			// Embed all the samples into a Euclidean space
			if (!embedder.embed(simMat, embeddedSamples))
			{
				cerr << "Error: Cannot embed similarity matrix" << endl;
				exit(EXIT_FAILURE);
			}
		}

		unsigned int n = embeddedSamples.getRowNum() - 1;	// Number of the training samples
		unsigned int d = embeddedSamples.getColumnNum();	// Dimension of the embedded vectors

		CvMat *_trainingData = cvCreateMat(n, d, CV_32FC1);	// Training data
		for (unsigned int i = 0; i < n; ++i)
		{
			for (unsigned int j = 0; j < d; ++j)
			{
				cvmSet(_trainingData, i, j, static_cast<float>(embeddedSamples(i, j)));
			}
		}

		CvMat *_testSample = cvCreateMat(1, d, CV_32FC1); // Test sample
		for (unsigned int j = 0; j < d; ++j)
		{
			cvmSet(_testSample, 0, j, static_cast<float>(embeddedSamples(n, j)));
		}

		CvMat *_labels = cvCreateMat(n, 1, CV_32SC1); // Class labels of the training data
		copy(labels.begin(), labels.end(), reinterpret_cast<int *>(cvPtr1D(_labels, 0)));

		// Set the SVM parameters
		CvSVMParams svmParams;
		svmParams.svm_type = CvSVM::C_SVC;
		svmParams.kernel_type = CvSVM::LINEAR;
		svmParams.C = parameters.C;
		svmParams.class_weights = NULL;
		svmParams.term_crit = cvTermCriteria(CV_TERMCRIT_ITER + CV_TERMCRIT_EPS, 3000, 1e-8);

		svm.train(_trainingData, _labels, NULL, NULL, svmParams); // Train the local SVM

		// Predict the class label of the test sample
		int result = static_cast<int>(svm.predict(_testSample));

		cvReleaseMat(&_trainingData);
		cvReleaseMat(&_testSample);
		cvReleaseMat(&_labels);

		return result;
	}

	// Check if the given labels belong to one class
	bool SVMKNN::isOneClass(const vector<int> &labels)
	{
		int y = labels.front();

		for (vector<int>::const_iterator itr = labels.begin();
			 itr != labels.end(); ++itr)
		{
			if (*itr != y) return false;
		}

		return true;
	}
}
