// File: SVMKNNTest.cpp
// Purpose: Implementation of the SVMKNNTest class
// Remarks: This class is designed as an adapter for the SVMKNN class
//			to enable testing on multiple sets of parameters.
// Author: Yihua Chen (yhchen@u.washington.edu)
// Last modified: 5/27/2008 4:21:38 PM

#include <cmath>
#include <iostream>
#include <algorithm>
#include "RealMatrix.h"
#include "SVMKNNTest.h"

using namespace std;

namespace idl
{
	const string SVMKNNTest::_ID = "SVM-KNN"; // Define the class ID

	// Register the class instantiator with the factory
	const bool SVMKNNTest::_REGISTERED = 
		ClassifierTestFactory::registerCreator(SVMKNNTest::_ID, &SVMKNNTest::create);

	// Destructor
	SVMKNNTest::~SVMKNNTest()
	{
		// Remarks: The string "factory trick" will never be printed.
		if (!_REGISTERED) cout << "factory trick";
	}

	// Configure the test for an SVM-KNN classifier
	bool SVMKNNTest::configure(ConfigParser &config, int secIdx)
	{
		// Get the embedding mode for SVM-KNN
		string strEmbeddingMode;
		if (!config.getString(secIdx, "EmbeddingMode", strEmbeddingMode))
		{
			cerr << "Error: Missing embedding mode for SVM-KNN classifier " << secIdx << endl;
			return false;
		}
		if ("shift" == strEmbeddingMode)
		{
			svmknn.setEmbeddingMode(SVMKNN::EMBED_SHIFT);
		}
		else if ("clip" == strEmbeddingMode)
		{
			svmknn.setEmbeddingMode(SVMKNN::EMBED_CLIP);
		}
		else if ("flip" == strEmbeddingMode)
		{
			svmknn.setEmbeddingMode(SVMKNN::EMBED_FLIP);
		}
		else if ("none" == strEmbeddingMode)
		{
			svmknn.setEmbeddingMode(SVMKNN::EMBED_NONE);
		}
		else
		{
			cerr << "Error: Unknown embedding mode for SVM-KNN classifier " << secIdx << endl;
			return false;
		}

		// Get the penalty parameters to cross-validate
		if (!config.getRealVec(secIdx, "C", Cs))
		{
			cerr << "Error: Missing or incorrect C for SVM-KNN classifier "
				 << secIdx << endl;
			return false;
		}

		// Check the retrieved penalty parameters for C-SVM
		if (!checkCs())
		{
			cerr << "Error: Incorrect C for SVM-KNN classifier " << secIdx << endl;
			return false;
		}

		// Get neighborhood sizes to cross-validate
		if (!config.getUIntVec(secIdx, "NeighborhoodSize", neighborhoodSizes))
		{
			cerr << "Error: Missing or incorrect neighborhood size for SVM-KNN classifier "
				 << secIdx << endl;
			return false;
		}

		// Check the retrieved neighborhood sizes
		if (!checkNeighborhoodSizes())
		{
			cerr << "Error: Incorrect neighborhood size for SVM-KNN classifier "
				 << secIdx << endl;
			return false;
		}

		reset(static_cast<unsigned int>(Cs.size() * neighborhoodSizes.size()));

		return true;
	}

	// Test the SVM-KNN classifier on all available parameters
	bool SVMKNNTest::test(const SimData &data, ofstream &logFile)
	{
		logFile << "SVM-KNN classifier" << endl
				<< "Embedding mode: " << svmknn.getEmbeddingName() << endl
				<< "Neighborhood size\tC\tError rate" << endl;

		unsigned int i = 0; // parameter set count

		vector<int> results(data.getTestSetSize());

		for (vector<unsigned int>::const_iterator itrNeighborhoodSize = neighborhoodSizes.begin();
			 itrNeighborhoodSize != neighborhoodSizes.end(); ++itrNeighborhoodSize)
		{
			svmknn.setNeighborhoodSize(*itrNeighborhoodSize);

			for (vector<double>::const_iterator itrC = Cs.begin(); itrC != Cs.end(); ++itrC)
			{
				svmknn.setRegPara(*itrC);
				if (!svmknn.classify(data, results)) return false;

				double errorRate = data.getClassificationError(results);

				// Log the error rate
				logFile << *itrNeighborhoodSize << "\t" << *itrC << "\t"
						<< (errorRate * 100.0) << "%" << endl;

				errorSum(i) += errorRate;
				errorSumOfSquares(i++) += (errorRate * errorRate);
			}
		}

		logFile << endl;

		++testCount;

		return true;
	}

	// Check the legitimacy of the penalty parameters for C-SVM
	bool SVMKNNTest::checkCs()
	{
		sort(Cs.begin(), Cs.end());

		return Cs.front() > 0.0; // All Cs must be greater than 0
	}

	// Check the legitimacy of the neighborhood sizes
	bool SVMKNNTest::checkNeighborhoodSizes()
	{
		sort(neighborhoodSizes.begin(), neighborhoodSizes.end());

		return neighborhoodSizes.front() > 0;
	}

	// Clear the results with the best parameters
	void SVMKNNTest::resetBestParaResults()
	{
		resultsBestPara.clear();
	}

	// Select the best parameters based on the previous tests
	void SVMKNNTest::selectBestPara()
	{
		ResultType latestResult;
		latestResult.testError = -1.0; // No test with the best parameters yet

		if (0 == testCount)
		{
			// No cross-validation performed
			latestResult.C = Cs.front();
			latestResult.neighborhoodSize = neighborhoodSizes.front();
			latestResult.cvError = 0.0;
			latestResult.cvErrorStdDev = 0.0;
		}
		else // testCount > 0
		{
			unsigned int i;

			// Get the minimum cross-validation error and its standard deviation
			getMinErrorInfo(i, latestResult.cvError, latestResult.cvErrorStdDev);

			// Get the neighborhood size corresponding to the minimum cross-validation error
			latestResult.neighborhoodSize = neighborhoodSizes[i / Cs.size()];

			// Get the C corresponding to the minimum cross-validation error
			latestResult.C = Cs[i % Cs.size()];
		}

		resultsBestPara.push_back(latestResult);
	}

	// Test the SVM-KNN classifier with the best parameters
	bool SVMKNNTest::testBestPara(const SimData &data, ofstream &logFile)
	{
		logFile << "SVM-KNN classifier" << endl
				<< "Embedding mode: " << svmknn.getEmbeddingName() << endl;

		unsigned int k = (resultsBestPara.back()).neighborhoodSize;
		svmknn.setNeighborhoodSize(k);

		double C = (resultsBestPara.back()).C;
		svmknn.setRegPara(C);

		vector<int> results(data.getTestSetSize());
		if (!svmknn.classify(data, results)) return false;

		double errorRate = data.getClassificationError(results, logFile);
		(resultsBestPara.back()).testError = errorRate;

		// Log the error rate
		logFile << "Neighborhood size\tC\tError rate" << endl
				<< k << "\t" << C << "\t" << (errorRate * 100.0) << "%" << endl << endl;

		return true;
	}

	// Report the test results of all available parameters
	void SVMKNNTest::reportResults(ofstream &file, bool verbose)
	{
		file << "SVM-KNN classifier" << endl
			 << "Embedding mode: " << svmknn.getEmbeddingName() << endl;

		if (verbose)
		{
			// Report the results of all the parameter sets
			file << "Neighborhood size\tC\tError rate\tStandard deviation" << endl;

			unsigned int i = 0;
			for (vector<unsigned int>::const_iterator itrNeighborhoodSize = neighborhoodSizes.begin();
				 itrNeighborhoodSize != neighborhoodSizes.end(); ++itrNeighborhoodSize)
			{
				for (vector<double>::const_iterator itrC = Cs.begin(); itrC != Cs.end(); ++itrC)
				{
					file << *itrNeighborhoodSize << "\t" << *itrC << "\t"
						 << (getError(i) * 100.0) << "%\t" << (getErrorStdDev(i) * 100.0) << "%" << endl;
					++i;
				}
			}
		}
		else
		{
			// Report the best result only
			file << "Neighborhood size: " << (resultsBestPara.back()).neighborhoodSize << endl
				 << "C: " << (resultsBestPara.back()).C << endl
				 << "Error rate: " << ((resultsBestPara.back()).cvError * 100.0) << "%" << endl
				 << "Standard deviation: " << ((resultsBestPara.back()).cvErrorStdDev * 100.0) << "%" << endl;
		}

		file << endl;
	}

	// Report the test results with the best parameters
	void SVMKNNTest::reportBestParaResults(ofstream &file, bool verbose)
	{
		file << "SVM-KNN classifier" << endl
			 << "Embedding mode: " << svmknn.getEmbeddingName() << endl;

		if (verbose)
		{
			file << "Neighborhood size: " << (resultsBestPara.back()).neighborhoodSize << endl
				 << "C: " << (resultsBestPara.back()).C << endl
				 << "Cross-validation error: " << ((resultsBestPara.back()).cvError * 100.0) << "%" << endl
				 << "Standard deviation: " << ((resultsBestPara.back()).cvErrorStdDev * 100.0) << "%" << endl
				 << "Test error: " << ((resultsBestPara.back()).testError * 100.0) << "%" << endl;
		}
		else
		{
			file << "Neighborhood size: " << (resultsBestPara.back()).neighborhoodSize << endl
				 << "C: " << (resultsBestPara.back()).C << endl
				 << "Error rate: " << ((resultsBestPara.back()).testError * 100.0) << "%" << endl;
		}

		file << endl;
	}

	// Summarize the results of all the tests with the best parameters
	void SVMKNNTest::reportBestParaSummary(ofstream &file, bool verbose)
	{
		file << "SVM-KNN classifier" << endl
			 << "Embedding mode: " << svmknn.getEmbeddingName() << endl;

		if (verbose)
		{
			file << "Neighborhood size\tC\tError rate" << endl;
		}

		unsigned int n = static_cast<unsigned int>(resultsBestPara.size());
		RealMatrix kVec(1, n);
		RealMatrix CVec(1, n);
		RealMatrix errorVec(1, n);

		unsigned int i = 0;
		for (vector<ResultType>::const_iterator itr = resultsBestPara.begin();
			 itr != resultsBestPara.end(); ++itr, ++i)
		{
			kVec(i) = static_cast<double>(itr->neighborhoodSize);
			CVec(i) = itr->C;
			errorVec(i) = itr->testError;

			if (verbose)
			{
				file << itr->neighborhoodSize << "\t" << itr->C << "\t"
					 << (itr->testError * 100.0) << "%" << endl;
			}
		}

		file << "Average neighborhood size: " << kVec.mean()
			 << " (" << kVec.std() << ")" << endl
			 << "Average C: " << CVec.mean()
			 << " (" << CVec.std() << ")" << endl
			 << "Average error rate: " << (errorVec.mean() * 100.0)
			 << "% (" << (errorVec.std() * 100.0) << "%)" << endl << endl;
	}
}
