// File: SimData.cpp
// Purpose: Implementation of the SimData class
// Remarks: This class is designed for classification data.
// Author: Yihua Chen (yhchen@u.washington.edu)
// Last modified: 7/6/2008 1:09:41 AM

#include <cstdlib>
#include <iostream>
#include <fstream>
#include <algorithm>
#include "SimData.h"

using namespace std;

namespace idl
{
	// Clear the current similarity data
	void SimData::clear()
	{
		numSamples = 0;
		numClasses = 0;

		normalizationFlag = false;
		symmetryFlag = false;
		simMat.clear();
		labels.clear();
		neighborIdx.clear();
		classMap.clear();
		
		loocvFlag = true;
		trainingSet.clear();
		testSet.clear();
		testNeighborIdx.clear();
	}

	// Reset the training and test settings
	void SimData::reset()
	{
		trainingSet.clear();
		testSet.clear();

		loocvFlag = true;
	}

	// Log data information
	void SimData::logDataInfo(ofstream &file) const
	{
		file << "Data set size: " << numSamples << endl
			 << "Normalized: " << (normalizationFlag ? "yes" : "no") << endl
			 << "Symmetric: " << (symmetryFlag ? "yes" : "no") << endl
			 << "Class number: " << classMap.size() << endl
			 << "Class distribution:" << endl;

		for (ClassMapType::const_iterator itr = classMap.begin(); itr != classMap.end(); ++itr)
		{
			unsigned int numMembers = static_cast<unsigned int>((itr->second).size());
			double percentage = static_cast<double>(numMembers) / static_cast<double>(numSamples);

			file << "Class " << itr->first << "\t" << numMembers << "\t"
				 << (percentage * 100.0) << "%" << endl;
		}
	}

	// Normalize similarity data to the range of [0, 1]
	bool SimData::normalizeData()
	{
		double maxSimVal = simMat.max();

		if (0.0 == maxSimVal)
		{
			cerr << "Error: All similarities are zero" << endl;
			return false;
		}

		simMat /= maxSimVal; // Normalize the similarities
		normalizationFlag = true;

		return true;
	}

	// Load similarity data and class labels
	bool SimData::loadData(const string &dataFileName,
						   const string &dataFormat,
						   const string &labelFileName,
						   bool  symmetrization)
	{
		clear(); // Clear the current data

		// Read similarity data
		if (!readDataFile(dataFileName, dataFormat)) return false;

		if (symmetrization && !symmetryFlag)
		{
			// Symmetrize the data
			for (unsigned int i = 0; i < numSamples; ++i)
			{
				for (unsigned int j = 0; j < i; ++j)
				{
					double v = (simMat(i, j) + simMat(j, i)) / 2;

					simMat(i, j) = v;
					simMat(j, i) = v;
				}
			}

			symmetryFlag = true;
		}

		// Read class labels
		if (!readLabelFile(labelFileName)) return false;

		buildNeighborIdx();	// Build the neighbor indices
		buildClassMap();	// Build the class map

		normalizationFlag = false;
		loocvFlag = true;

		return true;
	}

	// Read similarity data from file
	bool SimData::readDataFile(const string &fileName, const string &format)
	{
		ifstream dataFile(fileName.c_str());
		if (!dataFile)
		{
			cerr << "Error: Cannot open " << fileName << endl;
			return false;
		}

		if ("ascii" == format)
		{
			// Read data file in the ASCII format
			return readDataFileAscii(dataFile);
		}
		else if ("ali" == format)
		{
			// Read data file in Ali's format
			return readDataFileAli(dataFile);
		}
		else
		{
			cerr << "Error: Unknown data file format" << endl;
			return false;
		}
	}

	// Read similarity data file in the ASCII format
	bool SimData::readDataFileAscii(ifstream &dataFile)
	{
		dataFile >> numSamples; // Read sample number

		if (numSamples < 2) return false;

		// Set the size of the similarity matrix
		simMat.resize(numSamples, numSamples);

		// Read the similarities
		for (unsigned int i = 0; i < numSamples; ++i)
		{
			for (unsigned int j = 0; j < numSamples; ++j)
			{
				if (dataFile.eof()) return false;

				double simVal;
				dataFile >> simVal;
				if (simVal < 0.0)
				{
					cerr << "Error: Negative similarity value not allowed" << endl;
					return false;
				}
				simMat(i, j) = simVal;
			}
		}

		symmetryFlag = false; // Assume asymmetry of the data in the ASCII format

		return true;
	}

	// Read similarity data file in Ali's format
	bool SimData::readDataFileAli(ifstream &dataFile)
	{
		unsigned int m, n;

		dataFile >> m;
		dataFile >> n;

		if (m != n)
		{
			cerr << "Error: Similarity matrix is a rectangular matrix" << endl;
			return false;
		}
		
		numSamples = m;

		if (numSamples < 2) return false;

		// Initialize the similarity matrix
		simMat.zeros(numSamples);

		// Read the symmetry flag
		string strSymmetry;
		dataFile >> strSymmetry;
		if ("symmetric" == strSymmetry)
		{
			symmetryFlag = true;
		}
		else if ("asymmetric" == strSymmetry)
		{
			symmetryFlag = false;
		}
		else
		{
			cerr << "Error: Missing or incorrect symmetry flag" << endl;
			return false;
		}

		// Read the similarities
		while(!dataFile.eof())
		{
			unsigned int i, j;
			double v;

			dataFile >> i >> j >> v;

			if (!dataFile) break;

			if (i >= numSamples || j >= numSamples)
			{
				cerr << "Error: Matrix element index out of range" << endl;
				return false;
			}
			if (v < 0.0)
			{
				cerr << "Error: Negative similarity value not allowed" << endl;
				return false;
			}

			simMat(i, j) = v;
			if (symmetryFlag) simMat(j, i) = v;
		}

		return true;
	}

	// Read class labels from file
	bool SimData::readLabelFile(const string &fileName)
	{
		ifstream labelFile(fileName.c_str());
		if (!labelFile)
		{
			cerr << "Error: Cannot open " << fileName << endl;
			return false;
		}

		// Read the class labels
		for (unsigned int i = 0; i < numSamples; ++i)
		{
			if (labelFile.eof()) return false;

			int label;
			labelFile >> label;
			labels.push_back(label);
		}

		return true;
	}

	// Build the neighbor indices for all the samples
	void SimData::buildNeighborIdx()
	{
		neighborIdx.resize(numSamples, numSamples - 1);

		vector<_SortUnit> simVec(numSamples - 1);

		for (unsigned int i = 0; i < numSamples; ++i)
		{
			for (unsigned int j = 0, k = 0; j < numSamples; ++j)
			{
				if (j != i)
				{
					_SortUnit tmp;
					
					tmp.idx = j;
					tmp.val = simMat(i, j);
					simVec[k++] = tmp;
				}
			}

			stable_sort(simVec.begin(), simVec.end(), _greater);

			for (unsigned int j = 0; j < numSamples - 1; ++j)
			{
				neighborIdx(i, j) = simVec[j].idx;
			}
		}
	}

	// Build the map for class membership
	void SimData::buildClassMap()
	{
		numClasses = 0;

		for (unsigned int i = 0; i < numSamples; ++i)
		{
			int label = labels[i];

			if (classMap.find(label) == classMap.end())
			{
				++numClasses;
				vector<unsigned int> classMembers(1, i);
				classMap[label] = classMembers;
			}
			else
			{
				classMap[label].push_back(i);
			}
		}
	}

	// Set the training and test set
	bool SimData::setTrainingTest(const vector<unsigned int> &trainingIdx,
								  const vector<unsigned int> &testIdx)
	{
		unsigned int m = static_cast<unsigned int>(testIdx.size());
		unsigned int n = static_cast<unsigned int>(trainingIdx.size());

		if (m + n > numSamples || 0 == m || 0 == n)
		{
			cerr << "Error: Illegal training or test set" << endl;
			return false;
		}

		trainingSet = trainingIdx;
		testSet = testIdx;

		sort(trainingSet.begin(), trainingSet.end());
		sort(testSet.begin(), testSet.end());

		testNeighborIdx.resize(m, n);

		// Build the neighbor indices for the test samples
		for (unsigned int i = 0; i < m; ++i)
		{
			for (unsigned int j = 0, k = 0; j < numSamples - 1 && k < n; ++j)
			{
				unsigned int neighbor = neighborIdx(testSet[i], j);

				if (binary_search(trainingSet.begin(), trainingSet.end(), neighbor))
				{
					testNeighborIdx(i, k++) = neighbor;
				}
			}
		}

		trainingNeighborIdx.resize(n, n-1);

		// Build the neighbor indices for the training samples
		for (unsigned int i = 0; i < n; ++i)
		{
			for (unsigned int j = 0, k = 0; j < numSamples - 1 && k < n-1; ++j)
			{
				unsigned int neighbor = neighborIdx(trainingSet[i], j);

				if (binary_search(trainingSet.begin(), trainingSet.end(), neighbor) && i != neighbor)
				{
					trainingNeighborIdx(i, k++) = neighbor;
				}
			}
		}
		
		loocvFlag = false;

		return true;
	}

	// Set a subset of the samples for leave-one-out cross-validation
	bool SimData::setLoocvTest(const vector<unsigned int> &testIdx)
	{
		unsigned int m = static_cast<unsigned int>(testIdx.size());

		if (m > numSamples || m < 2)
		{
			cerr << "Error: Illegal test set" << endl;
			return false;
		}

		trainingSet.clear();
		testSet = testIdx;

		sort(testSet.begin(), testSet.end());

		testNeighborIdx.resize(m, m - 1);

		// Build the neighbor indices for this subset
		for (unsigned int i = 0; i < m; ++i)
		{
			for (unsigned int j = 0, k = 0; j < numSamples - 1 && k < m - 1; ++j)
			{
				unsigned int neighbor = neighborIdx(testSet[i], j);

				if (binary_search(testSet.begin(), testSet.end(), neighbor))
				{
					testNeighborIdx(i, k++) = neighbor;
				}
			}
		}

		loocvFlag = true;

		return true;
	}

	// Get the similarities between the nearest neighbors of a test sample
	void SimData::getNNSimMat(unsigned int testSampleIdx,
							  unsigned int neighborhoodSize,
							  RealMatrix   &nnSimMat) const
	{
		vector<unsigned int> neighbors(neighborhoodSize);

		const UIntMatrix &neighborTable = testSet.empty() ? neighborIdx : testNeighborIdx;

		for (unsigned int i = 0; i < neighborhoodSize; ++i)
		{
			neighbors[i] = neighborTable(testSampleIdx, i);
		}

		nnSimMat = simMat(neighbors, neighbors);
	}

	// Get the similarity matrix of a training sample and its nearest neighbors
	void SimData::getNNofTrainingSimMat(unsigned int sampleIdx,
							  unsigned int neighborhoodSize,
							  RealMatrix   &nnSimMat) const
	{
		vector<unsigned int> neighbors(neighborhoodSize);

		for (unsigned int i = 0; i < neighborhoodSize; ++i)
		{
			neighbors[i] = trainingNeighborIdx(sampleIdx, i);
		}

		nnSimMat = simMat(neighbors, neighbors);
	}
	
	// Get the similarity matrix of the test sample and its nearest neighbors
	void SimData::getNNTestSimMat(unsigned int testSampleIdx,
								  unsigned int neighborhoodSize,
								  RealMatrix   &nnTestSimMat) const
	{
		vector<unsigned int> samples(neighborhoodSize + 1);

		const UIntMatrix &neighborTable = testSet.empty() ? neighborIdx : testNeighborIdx;

		for (unsigned int i = 0; i < neighborhoodSize; ++i)
		{
			samples[i] = neighborTable(testSampleIdx, i);
		}

		samples[neighborhoodSize] = testSet.empty() ? testSampleIdx : testSet[testSampleIdx];

		nnTestSimMat = simMat(samples, samples);
	}

	void SimData::getNNofTrainingSimVec(unsigned int sampleIdx,
							  unsigned int neighborhoodSize,
							  RealMatrix   &nnSimVec) const
	{
		vector<unsigned int> neighbors(neighborhoodSize);

		for (unsigned int i = 0; i < neighborhoodSize; ++i)
		{
			neighbors[i] = trainingNeighborIdx(sampleIdx, i);
		}

		nnSimVec = simMat(sampleIdx, neighbors);
	}
	
	// Get the similarities between the test sample and its nearest neighbors
	void SimData::getNNSimVec(unsigned int testSampleIdx,
							  unsigned int neighborhoodSize,
							  RealMatrix   &nnSimVec) const
	{
		vector<unsigned int> neighbors(neighborhoodSize);

		const UIntMatrix &neighborTable = testSet.empty() ? neighborIdx : testNeighborIdx;

		for (unsigned int i = 0; i < neighborhoodSize; ++i)
		{
			neighbors[i] = neighborTable(testSampleIdx, i);
		}

		nnSimVec = simMat(testSet.empty() ? testSampleIdx : testSet[testSampleIdx], neighbors);
	}

	void SimData::getNNofTrainingLabels(unsigned int sampleIdx,
							  unsigned int neighborhoodSize,
							  vector<int>  &nnLabels) const
	{
		nnLabels.resize(neighborhoodSize);

		for (unsigned int i = 0; i < neighborhoodSize; ++i)
		{
			nnLabels[i] = labels[trainingNeighborIdx(sampleIdx, i)];
		}
	}
	
	// Get the labels of the nearest neighbors of a test sample
	void SimData::getNNLabels(unsigned int testSampleIdx,
							  unsigned int neighborhoodSize,
							  vector<int>  &nnLabels) const
	{
		nnLabels.resize(neighborhoodSize);

		const UIntMatrix &neighborTable = testSet.empty() ? neighborIdx : testNeighborIdx;

		for (unsigned int i = 0; i < neighborhoodSize; ++i)
		{
			nnLabels[i] = labels[neighborTable(testSampleIdx, i)];
		}
	}

	int SimData::get1NNofTrainingLabel(unsigned int sampleIdx) const
	{
		return labels[neighborIdx(sampleIdx, 0)];
			
	}
	
	// Get the label of the 1-nearest-neighbor of a test sample
	int SimData::get1NNLabel(unsigned int testSampleIdx) const
	{
		return testSet.empty() ? labels[neighborIdx(testSampleIdx, 0)]
			: labels[testNeighborIdx(testSampleIdx, 0)];
	}

	// Get the indices of the training samples
	void SimData::getTrainingSet(vector<unsigned int> &trainingIdx,
								 unsigned int testSampleIdx) const
	{
		if (trainingSet.empty())
		{
			if (testSet.empty())
			{
				trainingIdx.clear();

				for (unsigned int i = 0; i < testSampleIdx; ++i)
				{
					trainingIdx.push_back(i);
				}
				for (unsigned int i = testSampleIdx + 1; i < numSamples; ++i)
				{
					trainingIdx.push_back(i);
				}
			}
			else
			{
				trainingIdx = testSet;
				trainingIdx.erase(trainingIdx.begin() + testSampleIdx);
			}
		}
		else
		{
			trainingIdx = trainingSet;
		}
	}

	// Get the similarity matrix of the training samples
	void SimData::getTrainingSimMat(RealMatrix &trainingSimMat,
									unsigned int testSampleIdx) const
	{
		vector<unsigned int> trainingIdx;
		getTrainingSet(trainingIdx, testSampleIdx);

		trainingSimMat = simMat(trainingIdx, trainingIdx);
	}

	// Get the labels of the training samples
	void SimData::getTrainingLabels(vector<int> &trainingLabels,
									unsigned int testSampleIdx) const
	{
		vector<unsigned int> trainingIdx;
		getTrainingSet(trainingIdx, testSampleIdx);

		trainingLabels.clear();
		for (vector<unsigned int>::const_iterator itr = trainingIdx.begin();
			 itr != trainingIdx.end(); ++itr)
		{
			trainingLabels.push_back(labels[*itr]);
		}
	}

	// Get the similarity matrix of the training and test samples
	void SimData::getTrainingTestSimMat(RealMatrix &trainingTestSimMat,
										unsigned int testSampleIdx) const
	{
		vector<unsigned int> sampleIdx;
		getTrainingSet(sampleIdx, testSampleIdx);

		if (trainingSet.empty())
		{
			sampleIdx.push_back(testSet.empty() ? testSampleIdx : testSet[testSampleIdx]);
		}
		else
		{
			sampleIdx.insert(sampleIdx.end(), testSet.begin(), testSet.end());
		}

		trainingTestSimMat = simMat(sampleIdx, sampleIdx);
	}

	// Get the similarity matrix of the test sample and the training samples
	void SimData::getTestSimVec(unsigned int testSampleIdx, RealMatrix &testSimVec) const
	{
		vector<unsigned int> trainingIdx;
		getTrainingSet(trainingIdx, testSampleIdx);

		if (!testSet.empty())
		{
			testSampleIdx = testSet[testSampleIdx];
		}

		testSimVec = simMat(testSampleIdx, trainingIdx);
	}

	// Get the similarities of the test samples to the training samples
	void SimData::getTestSims(RealMatrix &testSims) const
	{
		if (!trainingSet.empty() && !testSet.empty())
		{
			testSims = simMat(testSet, trainingSet);
		}
	}

	// Get all unique similarity values
	void SimData::getSimDomain(RealMatrix &uniqueVals) const
	{
		simMat.unique(uniqueVals);
	}

	void SimData::getUniqueLabels(vector<int> &uniqueLabels) const
	{
		map<int,int> seen;
		
		// Find all unique labels and save as keys in "seen"
		for(vector<int>::const_iterator itr = labels.begin();
			itr != labels.end(); ++itr)
		{
			seen[*itr] = 1;
		}

		//save unique labels
		uniqueLabels.resize(seen.size());
		unsigned int i = 0;
		for (map<int,int>::iterator itr = seen.begin();
			itr != seen.end(); ++itr)
		{
			uniqueLabels[i++] = itr->first;
		}

		sort(uniqueLabels.begin(), uniqueLabels.end());
	}
	
	// Get the mean similarity of the test sample and it nearest neighbors of each class
	void SimData::getMeanSim(unsigned int testSampleIdx, vector<double> &meanSim,
							 vector<int> &classLabels, unsigned int neighborhoodSize) const
	{
		unsigned int numTrainingSamples = getTrainingSetSize();
		unsigned int numClassesDone = 0;
		unsigned int testSample = testSet.empty() ? testSampleIdx : testSet[testSampleIdx];
		const UIntMatrix &neighborTable = testSet.empty() ? neighborIdx : testNeighborIdx;

		map<int, double> sims;
		map<int, unsigned int> counts;

		for (unsigned int i = 0; i < numTrainingSamples && numClassesDone < numClasses; ++i)
		{
			unsigned int neighbor = neighborTable(testSampleIdx, i);
			int neighborLabel = labels[neighbor];

			if (counts.find(neighborLabel) == counts.end())
			{
				counts[neighborLabel] = 1;
				sims[neighborLabel] = simMat(testSample, neighbor);

				if (1 == neighborhoodSize) ++numClassesDone;
			}
			else if (counts[neighborLabel] < neighborhoodSize)
			{
				++counts[neighborLabel];
				sims[neighborLabel] += simMat(testSample, neighbor);

				if (counts[neighborLabel] == neighborhoodSize) ++numClassesDone;
			}
		}

		meanSim.clear();
		classLabels.clear();

		for (map<int, double>::const_iterator itr = sims.begin(); itr != sims.end(); ++itr)
		{
			meanSim.push_back(itr->second / static_cast<double>(counts[itr->first]));
			classLabels.push_back(itr->first);
		}
	}

	// Get the classification error
	double SimData::getClassificationError(vector<int> &estimatedLabels) const
	{
		unsigned int numTestSamples = getTestSetSize();

		if (estimatedLabels.size() != numTestSamples) 
		{
			cerr << "Error: Incorrect number of estimated labels" << endl;
			exit(EXIT_FAILURE);
		}

		unsigned int numErrors = 0;

		if (testSet.empty())
		{
			for (unsigned int i = 0; i < numTestSamples; ++i)
			{
				if (estimatedLabels[i] != labels[i]) ++numErrors;
			}
		}
		else
		{
			for (unsigned int i = 0; i < numTestSamples; ++i)
			{
				if (estimatedLabels[i] != labels[testSet[i]]) ++numErrors;
			}
		}

		double errorRate = static_cast<double>(numErrors);
		errorRate /= static_cast<double>(numTestSamples);

		return errorRate;
	}

	// Get the classification error and log the classification results
	double SimData::getClassificationError(vector<int> &estimatedLabels, ofstream &logFile) const
	{
		unsigned int numTestSamples = getTestSetSize();

		if (estimatedLabels.size() != numTestSamples)
		{
			cerr << "Error: Incorrect number of estimated labels" << endl;
			exit(EXIT_FAILURE);
		}

		logFile << "Test set size: " << numTestSamples << endl
				<< "Classification results:" << endl
				<< "Estimated class label\tTrue class label" << endl;

		unsigned int numErrors = 0;

		if (testSet.empty())
		{
			for (unsigned int i = 0; i < numTestSamples; ++i)
			{
				logFile << estimatedLabels[i] << '\t' << labels[i] << endl;

				if (estimatedLabels[i] != labels[i]) ++numErrors;
			}
		}
		else
		{
			for (unsigned int i = 0; i < numTestSamples; ++i)
			{
				logFile << estimatedLabels[i] << '\t' << labels[testSet[i]] << endl;

				if (estimatedLabels[i] != labels[testSet[i]]) ++numErrors;
			}
		}

		double errorRate = static_cast<double>(numErrors);
		errorRate /= static_cast<double>(numTestSamples);

		return errorRate;
	}
}
