// File: SimEmbed.h
// Purpose: Definition of the SimEmbed class
// Author: Yihua Chen (yhchen@u.washington.edu)
// Last modified: 3/5/2008 11:40:27 PM

#ifndef IDL_SIMEMBED_H
#define	IDL_SIMEMBED_H

#include "RealMatrix.h"

namespace idl
{
	/** Class for embedding samples in Euclidean space based on their similarities. */
	class SimEmbed
	{
	public:
		/** Enumeration type for spectrum transformation method. */
		enum SpectrumTransformType
		{
			SPECTRUM_SHIFT,	/**< spectrum shift: f(lambda) = lambda + C */
			SPECTRUM_CLIP,	/**< spectrum clip: f(lambda) = max(0, lambda) */
			SPECTRUM_FLIP	/**< spectrum flip: f(lambda) = abs(lambda) */
		};

		/** Default constructor. */
		SimEmbed() : compensationFlag(false), spectrumTransformMode(SPECTRUM_SHIFT) {}

		/**
		 * Constructor.
		 * @param[in]	t	spectrum transformation mode
		 */
		SimEmbed(SpectrumTransformType t) : compensationFlag(false), spectrumTransformMode(t) {}

		~SimEmbed() {}	/*< Destructor. */

		/**
		 * Set the flag for compensation in projection.
		 * @param[in]	flag	flag for compensation in projection
		 */
		void setCompensationFlag(bool flag)
		{
			compensationFlag = flag;
		}

		/**
		 * Set the spectrum transformation method for embedding.
		 * @param[in]	t	spectrum transformation mode
		 */
		void setSpectrumTransformMode(SpectrumTransformType t)
		{
			spectrumTransformMode = t;
		}

		/**
		 * Embed the similarity data into a Euclidean space.
		 * @param[in]	simMat				similarity matrix
		 * @param[out]	embeddedSamples		samples in the embedding space (each row vector is a sample)
		 * @return		true if success; false if failure
		 * @attention	The similarity matrix must be square.
		 */
		bool embed(const RealMatrix &simMat, RealMatrix &embeddedSamples);

		/**
		 * Project the similarity data into the embedding space.
		 * @param[in]	simMat				similarity matrix
		 * @param[out]	projectedSamples	samples in the embedding space (each row vector is a sample)
		 * @return		true if success; false if failure
		 * @remarks		The embedding space is determined by the latest call of embed().
		 */
		bool project(const RealMatrix &simMat, RealMatrix &projectedSamples) const;

	private:
		/**< Cutoff threshold for determining the embedding dimension. */
		static const double TH;

		/**< Flag indicating if the projection includes compensation or not. */
		bool compensationFlag;

		/** Spectrum transformation method for embedding. */
		SpectrumTransformType spectrumTransformMode;

		RealMatrix projMat; /**< Projection matrix. */
	};
}

#endif // IDL_SIMEMBED_H
