// File: SimTest.cpp
// Purpose: Implementation of the SimTest class
// Remarks: This class is designed for classification test.
// Author: Yihua Chen (yhchen@u.washington.edu)
// Last modified: 7/6/2008 1:10:15 AM

//#include <cstdlib>
//#include <ctime>
#include <cmath>
#include <limits>
#include <iostream>
#include <fstream>
#include <sstream>
#include <vector>
#include <algorithm>
#include "CmRandomNumberGenerator.h" // Mersenne Twister (singleton) header file
#include "utilities.h"
#include "ClassifierTest.h"
#include "SimTest.h"

using namespace std;

namespace idl
{
	// Constructor
	SimTest::SimTest()
		: testLevel(0), testMode(TEST_RANDOM), testDataPercentage(0.0), numTests(0),
		  cvMode(CV_NOCV), cvNfold(0), cvPercentage(0.0), numCVs(0)
	{
	}

	// Destructor
	SimTest::~SimTest()
	{
		// Delete all the classifier tests
		for (vector<ClassifierTest *>::iterator itr = classifierTests.begin();
			 itr != classifierTests.end(); ++itr)
		{
			delete *itr;
		}
	}

	// Clear the current settings
	void SimTest::clear()
	{
		dataFileName.clear();
		dataFormat.clear();
		labelFileName.clear();
		logFileName.clear();
		resultFileName.clear();
		partitionFileName.clear();

		testLevel = 0;
		testMode = TEST_RANDOM;
		testDataPercentage = 0.0;
		numTests = 0;
		cvMode = CV_NOCV;
		cvNfold = 0;
		cvPercentage = 0.0;
		numCVs = 0;

		data.clear();
		testSet.clear();
		trainingSet.clear();

		// Delete all the classifier tests
		for (vector<ClassifierTest *>::iterator itr = classifierTests.begin();
			 itr != classifierTests.end(); ++itr)
		{
			delete *itr;
		}
		classifierTests.clear();
	}

	// Configure the classification test
	bool SimTest::configure(ConfigParser &config)
	{
		// Get data file name
		if (!config.getString(-1, "DataFile", dataFileName))
		{
			cerr << "Error: Missing data file name" << endl;
			return false;
		}

		// Get data format
		if (!config.getString(-1, "DataFormat", dataFormat))
		{
			cerr << "Error: Missing data format" << endl;
			return false;
		}

		// Get label file name
		if (!config.getString(-1, "LabelFile", labelFileName))
		{
			cerr << "Error: Missing label file name" << endl;
			return false;
		}

		// Get data symmetrization flag
		string strDataSymmetrization;
		if (!config.getString(-1, "DataSymmetrization", strDataSymmetrization))
		{
			cerr << "Error: Missing data symmetrization flag" << endl;
			return false;
		}
		bool symmetrization;
		if ("yes" == strDataSymmetrization)
		{
			symmetrization = true;
		}
		else if ("no" == strDataSymmetrization)
		{
			symmetrization = false;
		}
		else
		{
			cerr << "Error: Unknown data symmetrization flag" << endl;
			return false;
		}

		// Load similarity data
		if (!data.loadData(dataFileName, dataFormat, labelFileName, symmetrization))
		{
			cerr << "Error: Cannot load similarity data" << endl;
			return false;
		}

		// Get data normalization flag
		string strDataNormalization;
		if (!config.getString(-1, "DataNormalization", strDataNormalization))
		{
			cerr << "Error: Missing data normalization flag" << endl;
			return false;
		}
		if ("yes" == strDataNormalization)
		{
			if (!data.normalizeData())
			{
				cerr << "Error: Cannot normalize similarity data" << endl;
				return false;
			}
		}
		else if ("no" != strDataNormalization)
		{
			cerr << "Error: Unknown data normalization flag" << endl;
			return false;
		}

		// Get log file name
		if (!config.getString(-1, "LogFile", logFileName))
		{
			cerr << "Error: Missing log file name" << endl;
			return false;
		}

		// Get result file name
		if (!config.getString(-1, "ResultFile", resultFileName))
		{
			cerr << "Error: Missing result file name" << endl;
			return false;
		}

		// Get partition file name
		if (!config.getString(-1, "PartitionFile", partitionFileName))
		{
			cerr << "Error: Missing partition file name" << endl;
			return false;
		}
		if ("none" == partitionFileName)
		{
			partitionFileName.clear();
		}

		// Get test level
		if (!config.getInt(-1, "TestLevel", testLevel))
		{
			cerr << "Error: Missing or unknown test level" << endl;
			return false;
		}
		if (1 != testLevel && 2 != testLevel)
		{
			cerr << "Error: Unknown test level" << endl;
			return false;
		}

		if (2 == testLevel)
		{
			// Get test mode
			string strTestMode;
			if (!config.getString(-1, "TestMode", strTestMode))
			{
				cerr << "Error: Missing or unknown test mode" << endl;
				return false;
			}
			if ("random" == strTestMode)
			{
				testMode = TEST_RANDOM;

				// Get test data percentage
				if (!config.getReal(-1, "TestDataPercentage", testDataPercentage))
				{
					cerr << "Error: Missing test data percentage" << endl;
					return false;
				}
				testDataPercentage /= 100.0;
				if (testDataPercentage <= 0.0 || testDataPercentage >= 1.0)
				{
					cerr << "Error: Incorrect test data percentage" << endl;
					return false;
				}

				// Get number of tests
				if (!config.getUInt(-1, "TestNumber", numTests))
				{
					cerr << "Error: Missing number of tests" << endl;
					return false;
				}
				if (0 == numTests)
				{
					cerr << "Error: Incorrect number of tests" << endl;
					return false;
				}
			}
			else if ("fixed" == strTestMode)
			{
				testMode = TEST_FIXED;

				// Get test set
				if (!config.getUIntVec(-1, "TestSet", testSet))
				{
					cerr << "Error: Missing or incorrect test set" << endl;
					return false;
				}
				sort(testSet.begin(), testSet.end());
				if(testSet.back() >= data.getDataSetSize())
				{
					cerr << "Error: Incorrect test set" << endl;
					return false;
				}
				getTrainingSet();
			}
			else if ("prespecified" == strTestMode)
			{
				testMode = TEST_PRESPECIFIED;

				if (partitionFileName.empty())
				{
					cerr << "Error: Missing partition file name" << endl;
					return false;
				}
			}
			else
			{
				cerr << "Error: Unknown test mode" << endl;
				return false;
			}
		}

		// Get cross-validation mode
		string strCVMode;
		if (!config.getString(-1, "CVMode", strCVMode))
		{
			cerr << "Error: Missing or unknown cross-validation mode" << endl;
			return false;
		}
		if ("nocv" == strCVMode)
		{
			if (1 == testLevel)
			{
				cerr << "Error: Incorrect cross-validation mode" << endl;
				return false;
			}
			else if (TEST_PRESPECIFIED == testMode) // 2 == testLevel
			{
				if (!config.getInt(-1, "TestSetIdx", testSetIdx))
				{
					cerr << "Error: Missing test set index" << endl;
					return false;
				}
				if (testSetIdx < -1)
				{
					cerr << "Error: Incorrect test set index" << endl;
					return false;
				}
			}
			cvMode = CV_NOCV;
		}
		else if ("loocv" == strCVMode)
		{
			cvMode = CV_LOOCV;
		}
		else if ("nfold" == strCVMode)
		{
			if (2 == testLevel && TEST_PRESPECIFIED == testMode)
			{
				cerr << "Error: Incorrect cross-validation mode" << endl;
				return false;
			}
			cvMode = CV_NFOLD;

			// Get n for n-fold cross-validation with random partitions
			if (!config.getUInt(-1, "CVnfold", cvNfold))
			{
				cerr << "Error: Missing n for n-fold cross-validation with random partitions" << endl;
				return false;
			}
			if (cvNfold < 2)
			{
				cerr << "Error: Incorrect n for n-fold cross-validation with random partitions" << endl;
				return false;
			}
		}
		else if ("random" == strCVMode)
		{
			if (2 == testLevel && TEST_PRESPECIFIED == testMode)
			{
				cerr << "Error: Incorrect cross-validation mode" << endl;
				return false;
			}
			cvMode = CV_RANDOM;

			// Get cross-validation test data percentage
			if (!config.getReal(-1, "CVPercentage", cvPercentage))
			{
				cerr << "Error: Missing cross-validation test data percentage" << endl;
				return false;
			}
			cvPercentage /= 100.0;
			if (cvPercentage <= 0.0 || cvPercentage >= 1.0)
			{
				cerr << "Error: Incorrect cross-validation test data percentage" << endl;
				return false;
			}

			// Get number of cross-validations
			if (!config.getUInt(-1, "CVNumber", numCVs))
			{
				cerr << "Error: Missing number of cross-validations" << endl;
				return false;
			}
			if (0 == numCVs)
			{
				cerr << "Error: Incorrect number of cross-validations" << endl;
				return false;
			}
		}
		else if ("prespecified" == strCVMode)
		{
			if (1 == testLevel)
			{
				if (!config.getInt(-1, "TestSetIdx", testSetIdx))
				{
					cerr << "Error: Missing test set index" << endl;
					return false;
				}
				if (testSetIdx < 0)
				{
					cerr << "Error: Incorrect test set index" << endl;
					return false;
				}

				int tmp;
				if (!config.getInt(-1, "CVSetIdx", tmp)) tmp = 0;
				if (-1 == tmp)
				{
					cvSetIdx.clear();
				}
				else
				{
					if (!config.getUIntVec(-1, "CVSetIdx", cvSetIdx))
					{
						cerr << "Error: Missing cross-validation set indices" << endl;
						return false;
					}
					sort(cvSetIdx.begin(), cvSetIdx.end());
				}
			}
			else if (TEST_PRESPECIFIED != testMode) // 2 == testLevel
			{
				cerr << "Error: Incorrect cross-validation mode" << endl;
				return false;
			}
			cvMode = CV_PRESPECIFIED;

			if (partitionFileName.empty())
			{
				cerr << "Error: Missing partition file name" << endl;
				return false;
			}
		}
		else
		{
			cerr << "Error: Unknown cross-validation mode" << endl;
			return false;
		}

		// Get configuration parameters for each classifier
		unsigned int numClassifiers = config.getClassifierNum();

		for (unsigned int i = 0; i < numClassifiers; ++i)
		{
			string strSwitch;

			if (!config.getString(i, "Switch", strSwitch))
			{
				cerr << "Error: Missing switch for classifier " << i << endl;
				return false;
			}
			if ("on" == strSwitch)
			{
				string strClassifier;

				if (!config.getString(i, "Classifier", strClassifier))
				{
					cerr << "Error: Missing classifier type for classifier " << i << endl;
					return false;
				}

				ClassifierTest *ptClassifierTest = ClassifierTestFactory::create(strClassifier);

				if (NULL == ptClassifierTest)
				{
					cerr << "Error: Unknown classifier type for classifier " << i << endl;
					return false;
				}

				if (!ptClassifierTest->configure(config, i)) return false;
				
				classifierTests.push_back(ptClassifierTest);
			}
			else if ("off" != strSwitch)
			{
				cerr << "Error: Unknown switch for classifier " << i << endl;
				return false;
			}
		}

		return true;
	}

	// Read the header of the partition file
	bool SimTest::readPartitionFileHeader()
	{
		ifstream partitionFile(partitionFileName.c_str());
		if (!partitionFile)
		{
			cerr << "Error: Cannot open partition file " << partitionFileName << endl;
			return false;
		}

		partitionFile >> numTests >> numCVs;
		if (!partitionFile || (2 == testLevel && 0 == numTests) ||
			(1 == testLevel && 0 == numCVs))
		{
			cerr << "Error: Incorrect or inappropriate partition file" << endl;
			return false;
		}

		return true;
	}

	// Write the header of the partition file
	bool SimTest::writePartitionFileHeader()
	{
		if (partitionFileName.empty() || (1 == testLevel && CV_LOOCV == cvMode)) return true;

		ofstream partitionFile(partitionFileName.c_str());
		if (!partitionFile)
		{
			cerr << "Error: Cannot open partition file " << partitionFileName << endl;
			return false;
		}

		// Write the number of tests
		if (1 == testLevel)
		{
			partitionFile << "0\t";
		}
		else if (TEST_RANDOM == testMode)
		{
			partitionFile << numTests << '\t';
		}
		else // TEST_FIXED == testMode
		{
			partitionFile << "1\t";
		}

		// Write the number of cross-validations per test
		if (CV_NOCV == cvMode || CV_LOOCV == cvMode)
		{
			partitionFile << '0' << endl;
		}
		else if (CV_NFOLD == cvMode)
		{
			partitionFile << cvNfold << endl;
		}
		else // CV_RANDOM == cvMode
		{
			partitionFile << numCVs << endl;
		}

		return true;
	}

	// Start the test
	bool SimTest::start(ConfigParser &config)
	{
		clear(); // Clear the current settings

		// Configure the classification test
		if (!configure(config)) return false;

		if ((1 == testLevel && CV_PRESPECIFIED == cvMode) ||
			(2 == testLevel && TEST_PRESPECIFIED == testMode))
		{
			// Read the header of the partition file
			if (!readPartitionFileHeader()) return false;

			if (0 == numCVs && CV_PRESPECIFIED == cvMode)
			{
				cvMode = CV_NOCV;
				testSetIdx = -1;
			}
		}
		else
		{
			// Write the header of the partition file
			if (!writePartitionFileHeader()) return false;
		}

		// Open log file
		ofstream logFile(logFileName.c_str());
		if (!logFile)
		{
			cerr << "Error: Cannot open log file " << logFileName << endl;
			return false;
		}

		// Open result file
		ofstream resultFile(resultFileName.c_str());
		if (!resultFile)
		{
			cerr << "Error: Cannot open result file " << resultFileName << endl;
			return false;
		}

		// Log test information (including data information)
		logTestInfo(logFile);
		logTestInfo(resultFile);

		if (1 == testLevel)
		{
			// Perform cross-validation only
			return crossValidate(logFile, resultFile, static_cast<unsigned int>(testSetIdx));
		}
		else if (TEST_FIXED == testMode)
		{
			// Perform test with fixed test set
			return singleTest(logFile, resultFile);
		}
		else // TEST_RANDOM == testMode || TEST_PRESPECIFIED == testMode
		{
			// Perform test with random or prespecified test set
			return multipleTest(logFile, resultFile);
		}
	}

	// Log test information
	void SimTest::logTestInfo(ofstream &file) const
	{
		if (1 == testLevel)
		{
			file << "Cross-Validation" << endl << endl;
		}
		else // 2 == testLevel
		{
			// Log test configurations
			if(TEST_RANDOM == testMode)
			{
				// Test with randomized training/test set
				file << "Random Test" << endl << endl
					 << "Test data percentage: " << (testDataPercentage * 100.0) << "%" << endl
					 << "Test number: " << numTests << endl;
			}
			else if (TEST_FIXED == testMode)
			{
				// Test with fixed test set
				file << "Fixed Test" << endl << endl
					 << "Test set size: " << testSet.size() << endl;
			}
			else // TEST_PRESPECIFIED == testMode
			{
				// Test with prespecified partitions
				file << "Prespecified Test" << endl << endl
					 << "Test number: " << ((CV_NOCV == cvMode && testSetIdx > -1) ? 1 : numTests) << endl;
			}
		}

		// Log cross-validation information
		file << "Cross-validation mode: ";
		switch (cvMode)
		{
		case CV_NOCV:
			file << "no cross-validation";
			break;
		case CV_LOOCV:
			file << "leave-one-out";
			break;
		case CV_NFOLD:
			file << "n-fold with random partitions" << endl << "Partition number: " << cvNfold;
			break;
		case CV_RANDOM:
			file << "random" << endl
				 << "Partition percentage: " << (cvPercentage * 100.0) << "%" << endl
				 << "Cross-validation number: " << numCVs;
			break;
		case CV_PRESPECIFIED:
			file << "prespecified partitions" << endl << "Cross-validation number: "
				 << ((1 == testLevel && !cvSetIdx.empty()) ? cvSetIdx.size(): numCVs);
			break;
		}
		file << endl << endl;

		// Log data information
		data.logDataInfo(file);

		file << endl;
	}

	// Perform multiple rounds of tests
	bool SimTest::multipleTest(ofstream &logFile, ofstream &resultFile)
	{
		if (TEST_PRESPECIFIED == testMode && CV_NOCV == cvMode && testSetIdx >= 0)
		{
			// Check test set index
			if (static_cast<unsigned int>(testSetIdx) >= numTests)
			{
				cerr << "Error: Incorrect test set index" << endl;
				return false;
			}

			// Load the specified test set from the partition file
			if (!loadTestSet(static_cast<unsigned int>(testSetIdx))) return false;

			// Perform test on the resulting training/test partition
			if (!singleTest(logFile, resultFile, static_cast<unsigned int>(testSetIdx))) return false;
		}
		else
		{
			for (unsigned int i = 0; i < numTests; ++i)
			{
				logFile << "--Test " << i << "----" << endl;
				resultFile << "--Test " << i << "----" << endl;

				if (TEST_RANDOM == testMode)
				{
					// Partition the data into the training and test sets randomly
					randomPartition(data.getDataSetSize(), testDataPercentage, testSet, trainingSet);
				}
				else // TEST_PRESPECIFIED == testMode
				{
					// Load the test set from the partition file
					if (!loadTestSet(i)) return false;
				}

				// Perform test on the resulting training/test partition
				if (!singleTest(logFile, resultFile, i)) return false;
			}

			logFile << "--Summary----" << endl;
			resultFile << "--Summary----" << endl;

			for (vector<ClassifierTest *>::const_iterator itr = classifierTests.begin();
				itr != classifierTests.end(); ++itr)
			{
				// Report the final results
				(*itr)->reportBestParaSummary(logFile, true);
				(*itr)->reportBestParaSummary(resultFile, false);
			}
		}

		return true;
	}

	// Perform a single round of test
	bool SimTest::singleTest(ofstream &logFile, ofstream &resultFile, unsigned int testIdx)
	{
		if (TEST_PRESPECIFIED != testMode)
		{
			// Save the test set to the partition file
			if (!saveTestSet()) return false;
		}

		// Perform cross-validation on the training set
		if (!crossValidate(logFile, resultFile, testIdx)) return false;

		data.setTrainingTest(trainingSet, testSet);

		for (vector<ClassifierTest *>::const_iterator itr = classifierTests.begin();
			 itr != classifierTests.end(); ++itr)
		{
			// Perform test with the best parameters
			if (!((*itr)->testBestPara(data, logFile))) return false;
		}

		logFile << "Test results" << endl << endl;
		resultFile << "Test results" << endl << endl;

		for (vector<ClassifierTest *>::const_iterator itr = classifierTests.begin();
			 itr != classifierTests.end(); ++itr)
		{
			// Report the final results
			(*itr)->reportBestParaResults(logFile, true);
			(*itr)->reportBestParaResults(resultFile, false);
		}

		return true;
	}

	// Perform cross-validation
	bool SimTest::crossValidate(ofstream &logFile, ofstream &resultFile, unsigned int testIdx)
	{
		// Clear the records of the previous tests
		for (vector<ClassifierTest *>::const_iterator itr = classifierTests.begin();
			 itr != classifierTests.end(); ++itr)
		{
			(*itr)->reset();
		}

		if (CV_LOOCV == cvMode)
		{
			// Leave-one-out cross-validation
			if (1 == testLevel)
			{
				// Leave-one-out cross-validation on the whole data set
				data.reset();
			}
			else
			{
				// Leave-one-out cross-validation on the training set only
				data.setLoocvTest(trainingSet);
			}
			for (vector<ClassifierTest *>::const_iterator itr = classifierTests.begin();
				 itr != classifierTests.end(); ++itr)
			{
				if (!((*itr)->test(data, logFile))) return false;
			}
		}
		else if (CV_NFOLD == cvMode || CV_PRESPECIFIED == cvMode)
		{
			// n-fold cross-validation with random or prespecified partitions
			vector< vector<unsigned int> > nfoldTestSets;
			vector< vector<unsigned int> > nfoldTrainingSets;

			if (CV_PRESPECIFIED == cvMode)
			{
				// Cross-validation with prespecified partitions
				if (1 == testLevel)
				{
					if (0 == numTests)
					{
						// Check test set index
						if (testIdx != 0)
						{
							cerr << "Error: Incorrect test set index" << endl;
							return false;
						}
					}
					else // numTests > 0
					{
						// Check test set index
						if (testIdx >= numTests)
						{
							cerr << "Error: Incorrect test set index" << endl;
							return false;
						}
						if (!loadTestSet(testIdx)) return false;
					}
				}

				// Load the specified cross-validation partitions from the partition file
				if (!loadPartitions(nfoldTestSets, nfoldTrainingSets, testIdx)) return false;

				if (1 == testLevel && !cvSetIdx.empty())
				{
					if (cvSetIdx.back() >= numCVs)
					{
						cerr << "Error: Incorrect cross-validation set indices" << endl;
						return false;
					}
					selectPartitions(nfoldTestSets, nfoldTrainingSets);
				}
			}
			else // CV_NFOLD == cvMode
			{
				// n-fold cross-validation with random partitions
				if (1 == testLevel)
				{
					// n-fold cross-validation on the whole data set
					nfoldPartition(data.getDataSetSize(), cvNfold, nfoldTestSets, nfoldTrainingSets);
				}
				else
				{
					// n-fold cross-validation on the training set only
					nfoldPartition(trainingSet, cvNfold, nfoldTestSets, nfoldTrainingSets);
				}

				if (!savePartitions(nfoldTestSets)) return false;
			}

			vector< vector<unsigned int> >::const_iterator itrTest;
			vector< vector<unsigned int> >::const_iterator itrTraining;

			for (itrTest = nfoldTestSets.begin(), itrTraining = nfoldTrainingSets.begin();
				 itrTest != nfoldTestSets.end() && itrTraining != nfoldTrainingSets.end();
				 ++itrTest, ++itrTraining)
			{
				data.setTrainingTest(*itrTraining, *itrTest);

				for (vector<ClassifierTest *>::const_iterator itr = classifierTests.begin();
					 itr != classifierTests.end(); ++itr)
				{
					if (!((*itr)->test(data, logFile))) return false;
				}
			}
		}
		else if (CV_RANDOM == cvMode)
		{
			// Randomized cross-validation
			vector<unsigned int> cvTestSet;
			vector<unsigned int> cvTrainingSet;

			for (unsigned int i = 0; i < numCVs; ++i)
			{
				if (1 == testLevel)
				{
					// Randomized cross-validation on the whole data set
					randomPartition(data.getDataSetSize(), cvPercentage, cvTestSet, cvTrainingSet);
				}
				else
				{
					// Randomized cross-validation on the training set only
					randomPartition(trainingSet, cvPercentage, cvTestSet, cvTrainingSet);
				}

				if (!savePartition(cvTestSet)) return false;

				data.setTrainingTest(cvTrainingSet, cvTestSet);

				for (vector<ClassifierTest *>::const_iterator itr = classifierTests.begin();
					 itr != classifierTests.end(); ++itr)
				{
					if (!((*itr)->test(data, logFile))) return false;
				}
			}
		}

		if (CV_NOCV != cvMode)
		{
			logFile << "Cross-validation results" << endl << endl;
			resultFile << "Cross-validation results" << endl << endl;
		}

		for (vector<ClassifierTest *>::const_iterator itr = classifierTests.begin();
			 itr != classifierTests.end(); ++itr)
		{
			// Choose the best parameters for each classifier
			(*itr)->selectBestPara();

			if (CV_NOCV != cvMode)
			{
				// Report the cross-validation results
				(*itr)->reportResults(logFile, true);		// Log the results in detail
				(*itr)->reportResults(resultFile, false);	// Report the best result only
			}
		}

		return true;
	}

	// Get the training set
	void SimTest::getTrainingSet()
	{
		trainingSet.clear();

		if (testSet.empty())
		{
			for (unsigned int i = 0; i < data.getDataSetSize(); ++i)
			{
				trainingSet.push_back(i);
			}
		}
		else
		{
			getComplement(data.getDataSetSize(), testSet, trainingSet);
		}
	}

	// Get the complement of a subset
	// Attention: The indices of the subset must be in ascending order.
	void SimTest::getComplement(unsigned int total,
								const vector<unsigned int> &subset,
								vector<unsigned int> &complement)
	{
		complement.clear();

		unsigned int p = 0;

		for (vector<unsigned int>::const_iterator itr = subset.begin();
			 itr != subset.end(); ++itr)
		{
			for (unsigned int i = p; i < *itr; ++i)
			{
				complement.push_back(i);
			}
			p = *itr + 1;
		}

		for (unsigned int i = p; i < total; ++i)
		{
			complement.push_back(i);
		}
	}

	// Get the relative complement of a set in another set
	// Attention: The indices of both set A and set B must be in ascending order.
	void SimTest::getComplement(const vector<unsigned int> &setA,
								const vector<unsigned int> &setB,
								vector<unsigned int> &complement)
	{
		complement.clear();

		vector<unsigned int>::const_iterator itrA = setA.begin();
		vector<unsigned int>::const_iterator itrB = setB.begin();
		while (itrA != setA.end() && itrB != setB.end())
		{
			if (*itrA < *itrB)
			{
				complement.push_back(*itrA);
				++itrA;
			}
			else // *itrA >= *itrB
			{
				if (*itrA == *itrB) ++itrA;
				++itrB;
			}
		}

		for (; itrA != setA.end(); ++itrA)
		{
			complement.push_back(*itrA);
		}
	}

	// Randomly divide the data into two parts
	void SimTest::randomPartition(unsigned int numSamples,
								  unsigned int numPartition1,
								  vector<unsigned int> &partition1,
								  vector<unsigned int> &partition2)
	{
		vector<bool> flags(numSamples, false);

		if (numPartition1 > numSamples)
		{
			numPartition1 = numSamples;
		}
		else if (numPartition1 < 1)
		{
			numPartition1 = 1;
		}

		//srand(static_cast<unsigned int>(time(NULL)));
		haruki::CmRandomNumberGenerator *pRandomNumberGenerator =
			haruki::CmRandomNumberGenerator::getInstance();

		for (unsigned int i = 0; i < numPartition1; )
		{
			//unsigned int idx = rand() % numSamples;
			unsigned int idx = pRandomNumberGenerator->getUInt() % numSamples;

			if (!flags[idx])
			{
				flags[idx] = true;
				++i;
			}
		}

		partition1.clear();
		partition2.clear();

		for (unsigned int i = 0; i < numSamples; ++i)
		{
			if (flags[i])
			{
				partition1.push_back(i);
			}
			else
			{
				partition2.push_back(i);
			}
		}
	}

	// Randomly divide the data into two parts
	void SimTest::randomPartition(const vector<unsigned int> &samples,
								  unsigned int numPartition1,
								  vector<unsigned int> &partition1,
								  vector<unsigned int> &partition2)
	{
		vector<unsigned int> partition1Idx;
		vector<unsigned int> partition2Idx;

		randomPartition(static_cast<unsigned int>(samples.size()), numPartition1, partition1Idx, partition2Idx);

		partition1.clear();

		for (vector<unsigned int>::const_iterator itr1 = partition1Idx.begin();
			 itr1 != partition1Idx.end(); ++itr1)
		{
			partition1.push_back(samples[*itr1]);
		}

		partition2.clear();

		for (vector<unsigned int>::const_iterator itr2 = partition2Idx.begin();
			 itr2 != partition2Idx.end(); ++itr2)
		{
			partition2.push_back(samples[*itr2]);
		}
	}

	// Randomly divide the data into two parts
	void SimTest::randomPartition(unsigned int numSamples,
								  double percentage,
								  vector<unsigned int> &partition1,
								  vector<unsigned int> &partition2)
	{
		unsigned int numPartition1 = static_cast<unsigned int>
			(floor(static_cast<double>(numSamples) * percentage + 0.5));

		randomPartition(numSamples, numPartition1, partition1, partition2);
	}

	// Randomly divide the data into two parts
	void SimTest::randomPartition(const vector<unsigned int> &samples,
								  double percentage,
								  vector<unsigned int> &partition1,
								  vector<unsigned int> &partition2)
	{
		unsigned int numPartition1 = static_cast<unsigned int>
			(floor(static_cast<double>(samples.size()) * percentage + 0.5));

		randomPartition(samples, numPartition1, partition1, partition2);
	}

	// Randomly divide the data into n equal parts
	void SimTest::nfoldPartition(unsigned int numSamples,
								 unsigned int n,
								 vector< vector<unsigned int> > &testSets,
								 vector< vector<unsigned int> > &trainingSets)
	{
		testSets.clear();
		trainingSets.clear();

		if (n > numSamples)
		{
			n = numSamples;
		}

		unsigned int m  = numSamples / n;
		unsigned int n2 = numSamples - m * n;	// number of partitions with (m+1) samples
		unsigned int n1 = n - n2;				// number of partitions with m samples

		vector<unsigned int> set0;
		vector<unsigned int> set1;
		vector<unsigned int> set2;
		vector<unsigned int> set3;

		for (unsigned int i = 0; i < numSamples; ++i)
		{
			set0.push_back(i);
		}

		for (unsigned int i = 0; i < n1; ++i)
		{
			randomPartition(set0, m, set1, set2);	// set0 = set1 + set2
			getComplement(numSamples, set1, set3);	// test/training = set1/set3

			testSets.push_back(set1);
			trainingSets.push_back(set3);

			set0 = set2;
		}

		for (unsigned int i = 0; i < n2; ++i)
		{
			randomPartition(set0, m + 1, set1, set2);	// set0 = set1 + set2
			getComplement(numSamples, set1, set3);		// test/training = set1/set3

			testSets.push_back(set1);
			trainingSets.push_back(set3);

			set0 = set2;
		}
	}

	// Randomly divide the data into n equal parts
	void SimTest::nfoldPartition(const vector<unsigned int> &samples,
								 unsigned int n,
								 vector< vector<unsigned int> > &testSets,
								 vector< vector<unsigned int> > &trainingSets)
	{
		vector< vector<unsigned int> > testSetsIdx;
		vector< vector<unsigned int> > trainingSetsIdx;

		nfoldPartition(static_cast<unsigned int>(samples.size()), n, testSetsIdx, trainingSetsIdx);

		_substitute(samples, testSetsIdx, testSets);
		_substitute(samples, trainingSetsIdx, trainingSets);
	}

	// Substitute the values for their indices
	void SimTest::_substitute(const vector<unsigned int> &samples,
							  const vector< vector<unsigned int> > &setsIdx,
							  vector< vector<unsigned int> > &sets)
	{
		sets.clear();

		vector<unsigned int> set;

		for (vector< vector<unsigned int> >::const_iterator itrSetsIdx = setsIdx.begin();
			 itrSetsIdx != setsIdx.end(); ++itrSetsIdx)
		{
			set.clear();

			for (vector<unsigned int>::const_iterator itrIdx = itrSetsIdx->begin();
				 itrIdx != itrSetsIdx->end(); ++itrIdx)
			{
				set.push_back(samples[*itrIdx]);
			}

			sets.push_back(set);
		}
	}

	// Skip lines in the input file stream
	void SimTest::_skipLines(ifstream &file, unsigned int nLines)
	{
		for (unsigned int i = 0; i < nLines; ++i)
		{
			file.ignore(numeric_limits<streamsize>::max(), '\n');
		}
	}

	// Load a test set from the partition file
	bool SimTest::loadTestSet(unsigned int testIdx)
	{
		ifstream partitionFile(partitionFileName.c_str());
		if (!partitionFile)
		{
			cerr << "Error: Cannot open partition file " << partitionFileName << endl;
			return false;
		}
		_skipLines(partitionFile, (1 + numCVs) * testIdx + 1);

		string line;
		getline(partitionFile, line);
		istringstream iss(line);

		testSet.clear();
		iss >> testSet;
		if (testSet.empty())
		{
			cerr << "Error: Cannot read test set " << testIdx << endl;
			return false;
		}
		sort(testSet.begin(), testSet.end());
		if(testSet.back() >= data.getDataSetSize())
		{
			cerr << "Error: Incorrect test set " << testIdx << endl;
			return false;
		}
		getTrainingSet();

		return true;
	}

	// Load cross-validation partitions from the partition file
	bool SimTest::loadPartitions(vector< vector<unsigned int> > &testSets,
								 vector< vector<unsigned int> > &trainingSets,
								 unsigned int testIdx) const
	{
		ifstream partitionFile(partitionFileName.c_str());
		if (!partitionFile)
		{
			cerr << "Error: Cannot open partition file " << partitionFileName << endl;
			return false;
		}
		_skipLines(partitionFile, (0 == numTests) ? 1 : ((1 + numCVs) * testIdx + 2));

		testSets.clear();
		trainingSets.clear();

		for (unsigned int i = 0; i < numCVs; ++i)
		{
			string line;
			getline(partitionFile, line);
			istringstream iss(line);

			vector<unsigned int> set0;	// test set
			vector<unsigned int> set1;	// training set

			iss >> set0;
			if (set0.empty())
			{
				cerr << "Error: Cannot read cross-validation partition " << testIdx << endl;
				return false;
			}
			sort(set0.begin(), set0.end());
			if(set0.back() >= data.getDataSetSize())
			{
				cerr << "Error: Incorrect cross-validation partition " << testIdx << endl;
				return false;
			}

			if (trainingSet.empty())
			{
				getComplement(data.getDataSetSize(), set0, set1);
			}
			else
			{
				getComplement(trainingSet, set0, set1);
			}

			testSets.push_back(set0);
			trainingSets.push_back(set1);
		}

		return true;
	}

	// Save cross-validation partitions to the partition file
	bool SimTest::savePartitions(const vector< vector<unsigned int> > &testSets) const
	{
		if (partitionFileName.empty()) return true;

		ofstream partitionFile(partitionFileName.c_str(), ios_base::app);
		if (!partitionFile)
		{
			cerr << "Error: Cannot open partition file " << partitionFileName << endl;
			return false;
		}

		for (vector< vector<unsigned int> >::const_iterator itr = testSets.begin();
			 itr != testSets.end(); ++itr)
		{
			partitionFile << *itr << endl;
		}

		return true;
	}

	// Save a cross-validation partition to the partition file
	bool SimTest::savePartition(const vector<unsigned int> &testIndices) const
	{
		if (partitionFileName.empty()) return true;

		ofstream partitionFile(partitionFileName.c_str(), ios_base::app);
		if (!partitionFile)
		{
			cerr << "Error: Cannot open partition file " << partitionFileName << endl;
			return false;
		}

		partitionFile << testIndices << endl;

		return true;
	}

	// Save the test set to the partition file
	bool SimTest::saveTestSet() const
	{
		return savePartition(testSet);
	}

	// Select a subset of the cross-validation partitions
	void SimTest::selectPartitions(vector< vector<unsigned int> > &testSets,
								   vector< vector<unsigned int> > &trainingSets) const
	{
		vector<bool> flags(numCVs, false);

		// Set the flags for the selected subset
		for (vector<unsigned int>::const_iterator itr = cvSetIdx.begin(); itr != cvSetIdx.end(); ++itr)
		{
			flags[*itr] = true;
		}

		// Remove the test and training sets not selected
		vector< vector<unsigned int> >::iterator itrTestSets = testSets.begin();
		vector< vector<unsigned int> >::iterator itrTrainingSets = trainingSets.begin();
		for (vector<bool>::const_iterator itrFlags = flags.begin(); itrFlags != flags.end(); ++itrFlags)
		{
			if (*itrFlags)
			{
				// Test and training sets selected
				++itrTestSets;
				++itrTrainingSets;
			}
			else
			{
				// Test and training sets not selected
				itrTestSets = testSets.erase(itrTestSets);
				itrTrainingSets = trainingSets.erase(itrTrainingSets);
			}
		}
	}
}
