// File: SimWeights.h
// Purpose: Definition of the SimWeights class
// Author: Yihua Chen (yhchen@u.washington.edu)
// Last modified: 6/2/2008 11:45:18 AM

#ifndef IDL_SIMWEIGHTS_H
#define IDL_SIMWEIGHTS_H

#include <vector>
#include "RealMatrix.h"

namespace idl
{
	/** Class for computing the sample weights based on the similarities. */
	class SimWeights
	{
	public:
		/** Enumeration type for Gram matrix approximation method. */
		enum ApproxType
		{
			APPROX_EIG_SHIFT,	/**< eigenvalue shift */
			APPROX_EIG_CLIP,	/**< eigenvalue clip */
			APPROX_EIG_FLIP,	/**< eigenvalue flip */
			APPROX_NONE			/**< no approximation */
		};

		SimWeights();	/**< Default constructor. */

		/**
		 * Constructor.
		 * @param[in]	n	number of variables
		 */
		SimWeights(unsigned int n);

		~SimWeights();	/*< Destructor. */

		/**
		 * Compute the similarity weights.
		 * @param[in]	simVec		vector of the similarities between the test sample
		 *							and its neighbors
		 * @param[out]	weights		vector of the weights
		 */
		void computeSimWeights(const RealMatrix		&simVec,
							   std::vector<double>	&weights);

		/**
		 * Compute the Markov weights.
		 * @param[in]	simMat		matrix of the similarities between the neighbors
		 * @param[in]	simVec		vector of the similarities between the test sample
		 *							and its neighbors
		 * @param[out]	weights		vector of the weights
		 * @param[in]	lambda		regularization parameter
		 * @param[in]	dissim		flag for diversity policy: true for encouraging
		 *							diversity; false for suppressing diversity (true
		 *							by default)
		 * @return		true if success; false if failure
		 * @attention	The similarities must be within the range of [0, 1].
		 */
		bool computeMarkovWeights(const RealMatrix		&simMat,
								  const RealMatrix		&simVec,
								  std::vector<double>	&weights,
								  double				lambda,
								  bool					dissim = true);

		/**
		 * Compute the sim-LIME weights.
		 * @param[in]	simMat		matrix of the similarities between the neighbors
		 * @param[in]	simVec		vector of the similarities between the test sample
		 *							and its neighbors
		 * @param[out]	weights		vector of the weights
		 * @param[in]	lambda		regularization parameter
		 * @param[in]	approx		method to use a similarity matrix to approximate
		 *							a Gram matrix (eigenvalue shift by default)
		 * @param[in]	symFlag		flag indicating if the similarity is symmetirc: true
		 *							if symmetric; false if asymmetric (false by default)
		 * @return		true if success; false if failure
		 */
		bool computeLimeWeights(const RealMatrix		&simMat,
								const RealMatrix		&simVec,
								std::vector<double>		&weights,
								double					lambda,
								ApproxType				approx = APPROX_EIG_SHIFT,
								bool					symFlag = false);

		/**
		 * Compute the sim-LIMV weights.
		 * @param[in]	simMat		matrix of the similarities between the neighbors
		 * @param[in]	simVec		vector of the similarities between the test sample
		 *							and its neighbors
		 * @param[out]	weights		vector of the weights
		 * @param[in]	lambda		regularization parameter
		 * @param[in]	approx		method to use a similarity matrix to approximate
		 *							a Gram matrix (eigenvalue shift by default)
		 * @param[in]	symFlag		flag indicating if the similarity is symmetirc: true
		 *							if symmetric; false if asymmetric (false by default)
		 * @return		true if success; false if failure
		 */
		bool computeLimvWeights(const RealMatrix		&simMat,
								const RealMatrix		&simVec,
								std::vector<double>		&weights,
								double					lambda,
								ApproxType				approx = APPROX_EIG_SHIFT,
								bool					symFlag = false);

		/**
		 * Compute the linear regression weights.
		 * @param[in]	simMat		matrix of the similarities between the neighbors
		 * @param[in]	simVec		vector of the similarities between the test sample
		 *							and its neighbors
		 * @param[out]	weights		vector of the weights
		 * @param[in]	lambda		regularization parameter
		 * @param[in]	approx		method to use a similarity matrix to approximate
		 *							a Gram matrix (no approximation by default)
		 * @param[in]	symFlag		flag indicating if the similarity is symmetirc: true
		 *							if symmetric; false if asymmetric (false by default)
		 * @return		true if success; false if failure
		 */
		bool computeLinearRegressionWeights(const RealMatrix		&simMat,
											const RealMatrix		&simVec,
											std::vector<double>		&weights,
											double					lambda,
											ApproxType				approx = APPROX_NONE,
											bool					symFlag = false);

	private:
		/**
		 * Resize the buffer.
		 * @param[in]	k	number of the weights
		 */
		void resizeBuffer(unsigned int k);

		/**
		 * Generate matrix Q.
		 * @param[in]	simMat		similarity matrix
		 * @param[in]	approx		method for Gram matrix approximation
		 * @param[in]	symFlag		flag indicating if the similarity matrix is symmetric:
		 *							true if symmetric; false if asymmetric (false by default)
		 * @remarks		This function assumes that the buffer for matrix Q is big enough.
		 */
		void generateQMtx(const RealMatrix &simMat, ApproxType approx, bool symFlag = false);

	private:
		static const double EPS; /**< Small positive number for regularization. */

		unsigned int nVariables; /**< Number of variables. */

		double *QMtx;	/**< Pointer to the buffer for matrix Q. */
		double *qVec;	/**< Pointer to the buffer for vector q. */
		double *wVec;	/**< Pointer to the buffer for the weights. */
	};
}

#endif // IDL_SIMWEIGHTS_H
