// GNU General Public License Agreement
// Copyright (C) 2004-2007 CodeCogs, Zyba Ltd, Broadwood, Holford, TA5 1DU, England.
//
// This program is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by CodeCogs. 
// You must retain a copy of this licence in all copies. 
//
// This program is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
// PARTICULAR PURPOSE. See the GNU General Public License for more details.
// ---------------------------------------------------------------------------------
//! Calculates the minimum of a one-dimensional real function using Brent's method.

#ifndef IDL_BRENT_H
#define IDL_OPTIMIZATION_BRENT_H

#define RGOLD       0.3819660
#define ZEPS        1.0E-10
#define MIN(x, y)   ((x) < (y) ? (x) : (y))
#define MAX(x, y)   ((x) > (y) ? (x) : (y))
#define ABS(x)      ((x) < 0 ? -(x) : (x))
#define SIGN(x, y)  ((y) < 0 ? -ABS(x) : ABS(x))

namespace idl
{

//! Calculates the minimum of a one-dimensional real function using Brent's method.

double brent (double (*f)(double, const RealMatrix&,double), const RealMatrix &domainVector, double classMean, double a, double b, double c, double eps = 1E-10, int maxit = 1000)
{

    double ax = MIN(a, c), bx = MAX(a, c), u, v = b,
    w = v, x = v, d = 0, e = 0, fx = f(x,domainVector,classMean), fv = fx, fw = fx;

    for (int i = 0; i < maxit; i++) {

        bool flag = true;
        double xm = (ax + bx) / 2, eps1 = eps * ABS(x) + ZEPS, eps2 = 2 * eps1;

        if (ABS(x - xm) <= eps2 - (bx - ax) / 2) break;

        if (ABS(e) > eps1) {

            double r = (x - w) * (fx - fv), q = (x - v) * (fx - fw),
            p = (x - v) * q - (x - w) * r, etemp = e;

            q = (q - r) / 5;
            if (q > 0) p *= -1;
            q = ABS(q), e = d;

            if (ABS(p) < ABS(q * etemp / 2) && p > q * (ax - x) && p < q * (bx - x)) {
                flag = false, d = p / q, u = x + d;
                if (u - ax < eps2 || bx - u < eps2) d = SIGN(eps1, xm - x);
            }

        }

        if (flag) (e = ((x < xm) ? bx : ax) - x);

        d = RGOLD * e;
        u = x + (ABS(d) < eps1 ? SIGN(eps1, d) : d);

        double fu = f(u, domainVector,classMean);
        if (fu <= fx) {

            (u < x ? bx : ax) = x;
            v = w, fv = fw;
            w = x, fw = fx;
            x = u, fx = fu;

        }
        else {

            (u < x ? ax : bx) = u;

            if (fu <= fw || w == x)
                v = w, fv = fw, w = u, fw = fu;
            else
            if (fu <= fv || v == x || v == w)
                v = u, fv = fu;

        }
    }

    return x;

}

}

#undef SIGN
#undef ABS
#undef MAX
#undef MIN
#undef ZEPS
#undef RGOLD

#endif

