// File: utilities.h
// Purpose: Definition of general utility functions and macros
// Author: Yihua Chen (yhchen@u.washington.edu)
// Last modified: 4/11/2008 6:36:17 PM

#ifndef IDL_UTILITIES_H
#define IDL_UTILITIES_H

#include <istream>
#include <ostream>
#include <string>
#include <sstream>
#include <vector>
#include "Matrix.h"

namespace idl
{
	/**
	 * Operator template for reading a vector from an input stream.
	 * @param[in]	istr	input stream
	 * @param[out]	vec		vector
	 * @return		reference to the input stream
	 * @attention	The new elements read from the input stream are
	 *				appended to the end of the vector. The original
	 *				elements of the vector are not erased.
	 */
	template <class T>
	std::istream & operator >> (std::istream &istr, std::vector<T> &vec)
	{
		T val;
		while (!istr.eof())
		{
			istr >> val;
			if (!istr) break;
			vec.push_back(val);
		}
		return istr;
	}

	/**
	 * Operator template for writing a vector to an output stream.
	 * @param[out]	ostr	output stream
	 * @param[in]	vec		vector
	 * @return		reference to the output stream
	 */
	template <class T>
	std::ostream & operator << (std::ostream &ostr, const std::vector<T> &vec)
	{
		for (typename std::vector<T>::const_iterator itr = vec.begin();
			 itr != vec.end(); ++itr)
		{
			ostr << *itr << ' ';
		}
		return ostr;
	}

	/**
	 * Operator template for writing a matrix to an output stream.
	 * @param[out]	ostr	output stream
	 * @param[in]	mat		matrix
	 * @return		reference to the output stream
	 */
	template <class T>
	std::ostream & operator << (std::ostream &ostr, const Matrix<T> &mat)
	{
		unsigned int m = mat.getRowNum();
		unsigned int n = mat.getColumnNum();

		for (unsigned int i = 0; i < m; ++i)
		{
			for (unsigned int j = 0; j < n - 1; ++j)
			{
				ostr << mat(i, j) << ' ';
			}
			ostr << mat(i, n - 1) << std::endl;
		}

		return ostr;
	}

	/**
	 * Convert a string to the value of a specific type.
	 * @param[in]	str		string to convert
	 * @return		value of a specific type converted from the given string
	 * @attention	The result is undefined if the conversion fails.
	 * @remarks		No exception will be thrown when the conversion fails.
	 */
	template <class T>
	inline T from_str(std::string &str)
	{
		std::istringstream iss(str);
		T val;
		iss >> val;
		return val;
	}

	/**
	 * Convert a value of a specific type to a string.
	 * @param[in]	val		value of a specific type to convert
	 * @return		string converted from the given value
	 * @attention	The result is undefined if the conversion fails.
	 * @remarks		No exception will be thrown when the conversion fails.
	 */
	template <class T>
	inline std::string to_str(T &val)
	{
		std::ostringstream oss;
		oss << val;
		return oss.str();
	}
}

#endif // IDL_UTILITIES_H
